from __future__ import annotations

from ctypes import (
    POINTER,
    POINTER as C_POINTER,
    Structure,
    byref,
    c_bool,
    c_char_p,
    c_int,
    c_uint,
    c_ulong,
    c_ulonglong,
    c_void_p,
    c_wchar_p,
    windll,
)
from ctypes.wintypes import BOOL, DWORD, HWND, LPCWSTR, LPWSTR, ULONG
from enum import IntFlag
from typing import TYPE_CHECKING, Callable, ClassVar, Sequence

import comtypes  # pyright: ignore[reportMissingTypeStubs]

from comtypes import COMMETHOD  # pyright: ignore[reportMissingTypeStubs]

from utility.system.win32.com.com_types import GUID
from utility.system.win32.hresult import HRESULT, S_OK  # pyright: ignore[reportMissingTypeStubs]

if TYPE_CHECKING:
    import os

    from ctypes import Array, _CData, _FuncPointer, _Pointer

    from comtypes._memberspec import _ComMemberSpec  # pyright: ignore[reportMissingTypeStubs]
    from typing_extensions import Self

# GUID Definitions
IID_IUnknown = GUID("{00000000-0000-0000-C000-000000000046}")
IID_IDispatch = GUID("{00020400-0000-0000-C000-000000000046}")
IID_IClassFactory = GUID("{00000001-0000-0000-C000-000000000046}")
IID_IStream = GUID("{0000000c-0000-0000-C000-000000000046}")
IID_IStorage = GUID("{0000000b-0000-0000-C000-000000000046}")
IID_IBindCtx = GUID("{0000000e-0000-0000-C000-000000000046}")
IID_IEnumShellItems = GUID("{70629033-E363-4A28-A567-0DB78006E6D7}")
IID_IContextMenu = GUID("{000214e4-0000-0000-c000-000000000046}")
IID_IContextMenu2 = GUID("{000214f4-0000-0000-c000-000000000046}")
IID_IContextMenu3 = GUID("{bcfce0a0-ec17-11d0-8d10-00a0c90f2719}")
IID_IShellFolder = GUID("{000214E6-0000-0000-C000-000000000046}")
IID_IShellFolder2 = GUID("{93F2F68C-1D1B-11D3-A30E-00C04F79ABD1}")
IID_IShellItem = GUID("{43826D1E-E718-42EE-BC55-A1E261C37BFE}")
IID_IShellItem2 = GUID("{7E9FB0D3-919F-4307-AB2E-9B1860310C93}")
IID_IShellLibrary = GUID("{11A66EFA-382E-451A-9234-1E0E12EF3085}")
IID_IShellItemArray = GUID("{B63EA76D-1F85-456F-A19C-48159EFA858B}")
IID_IShellItemFilter = GUID("{2659B475-EEB8-48B7-8F07-B378810F48CF}")
IID_IShellView = GUID("{000214e3-0000-0000-c000-000000000046}")
IID_IModalWindow = GUID("{B4DB1657-70D7-485E-8E3E-6FCB5A5C1802}")
IID_IFileDialog = GUID("{42F85136-DB7E-439C-85F1-E4075D135FC8}")
IID_IFileDialog2 = GUID("{61744FC7-85B5-4791-A9B0-272276309B13}")
IID_IFileSaveDialog = GUID("{84BCCD23-5FDE-4CDB-AEA4-AF64B83D78AB}")
IID_IFileSaveDialogOld = GUID("{2804B74C-AC16-4398-9DC0-DB83F5B7ED14}")
IID_IFileSaveDialogPrivate = GUID("{6CB95A6A-88B6-4DC4-B3EA-3A776D1E8EFF}")
IID_IFileOpenDialog = GUID("{D57C7288-D4AD-4768-BE02-9D969532D960}")
IID_IFileDialogEvents = GUID("{973510DB-7D7F-452B-8975-74A85828D354}")
IID_FileDialogPermissionAttribute = GUID("{0CCCA629-440F-313E-96CD-BA1B4B4997F7}")
IID_FileDialogPermission = GUID("{A8B7138C-8932-3D78-A585-A91569C743AC}")
IID_IFileDialogPrivate = GUID("{9EA5491C-89C8-4BEF-93D3-7F665FB82A33}")
IID_IFileDialogCustomize = GUID("{E6FDD21A-163F-4975-9C8C-A69F1BA37034}")
IID_IFileDialogEventsPrivate = GUID("{050E9E69-BAEA-4C08-AD6A-61666DD32E96}")
IID_IFileDialogControlEvents = GUID("{36116642-D713-4B97-9B83-7484A9D00433}")
IID_IFileDialogResultHandler = GUID("{42841501-194F-478F-9B4C-78985419DA53}")
IID_IShellLink = GUID("{000214f9-0000-0000-c000-000000000046}")
IID_IShellLinkDataList = GUID("{45E2B4AE-B1C3-11D0-BA91-00C04FD7A083}")
IID_IPropertyStore = GUID("{886D8EEB-8CF2-4446-8D02-CDBA1DBDCF99}")
IID_IFileOperationProgressSink = GUID("{04B0F1A7-9490-44BC-96E1-4296A31252E2}")
IID_IFileOperation = GUID("{94EA2B94-E9CC-49E0-C0E3-D20A7D91AA98}")
CLSID_FileOperation = GUID("{3AD05575-8857-4850-9277-11B85BDB8E09}")
CLSID_FileDialog = GUID("{3D9C8F03-50D4-4E40-BB11-70E74D3F10F3}")
CLSID_FileOpenDialog = GUID("{DC1C5A9C-E88A-4dde-A5A1-60F82A20AEF7}")
CLSID_FileOpenDialogLegacy = GUID("{725F645B-EAED-4fc5-B1C5-D9AD0ACCBA5E}")
CLSID_FileSaveDialog = GUID("{C0B4E2F3-BA21-4773-8DBA-335EC946EB8B}")
CLSID_FileSaveDialogLegacy = GUID("{AF02484C-A0A9-4669-9051-058AB12B9195}")
CLSID_ShellItemArrayShellNamespacehelper = GUID("{26671179-2ec2-42bf-93d3-64108589cad5}")
CLSID_ShellItemArrayShellNamespacehelper = GUID("{b77b1cbf-e827-44a9-a33a-6ccfeeaa142a}")  # redef??
CLSID_ShellItemArrayShellNamespacehelper = GUID("{CDC82860-468D-4d4e-B7E7-C298FF23AB2C}")  # redef??
CLSID_ShellItemArrayShellNamespacehelper = GUID("{F6166DAD-D3BE-4ebd-8419-9B5EAD8D0EC7}")  # redef??
CLSID_ShellLibraryAPI = GUID("{d9b3211d-e57f-4426-aaef-30a806add397}")
CLSID_ShellFileSystemFolder = GUID("{F3364BA0-65B9-11CE-A9BA-00AA004AE837}")
CLSID_ShellBindStatusCallbackProxy = GUID("{2B4F54B1-3D6D-11d0-8258-00C04FD5AE38}")
CLSID_ShellURL = GUID("{4bec2015-bfa1-42fa-9c0c-59431bbe880e}")
CLSID_ShellDropTarget = GUID("{4bf684f8-3d29-4403-810d-494e72c4291b}")
CLSID_ShellNameSpace = GUID("{55136805-B2DE-11D1-B9F2-00A0C98BC547}")


# Constants
class FileOpenOptions(IntFlag):
    FOS_UNKNOWN1 = 0x00000001
    FOS_OVERWRITEPROMPT = 0x00000002
    FOS_STRICTFILETYPES = 0x00000004
    FOS_NOCHANGEDIR = 0x00000008
    FOS_UNKNOWN2 = 0x00000010
    FOS_PICKFOLDERS = 0x00000020
    FOS_FORCEFILESYSTEM = 0x00000040
    FOS_ALLNONSTORAGEITEMS = 0x00000080
    FOS_NOVALIDATE = 0x00000100
    FOS_ALLOWMULTISELECT = 0x00000200
    FOS_UNKNOWN4 = 0x00000400
    FOS_PATHMUSTEXIST = 0x00000800
    FOS_FILEMUSTEXIST = 0x00001000
    FOS_CREATEPROMPT = 0x00002000
    FOS_SHAREAWARE = 0x00004000
    FOS_NOREADONLYRETURN = 0x00008000
    FOS_NOTESTFILECREATE = 0x00010000
    FOS_HIDEMRUPLACES = 0x00020000
    FOS_HIDEPINNEDPLACES = 0x00040000
    FOS_UNKNOWN5 = 0x00080000
    FOS_NODEREFERENCELINKS = 0x00100000
    FOS_UNKNOWN6 = 0x00200000
    FOS_UNKNOWN7 = 0x00400000
    FOS_UNKNOWN8 = 0x00800000
    FOS_UNKNOWN9 = 0x01000000
    FOS_DONTADDTORECENT = 0x02000000
    FOS_UNKNOWN10 = 0x04000000
    FOS_UNKNOWN11 = 0x08000000
    FOS_FORCESHOWHIDDEN = 0x10000000
    FOS_DEFAULTNOMINIMODE = 0x20000000
    FOS_FORCEPREVIEWPANEON = 0x40000000
    FOS_UNKNOWN12 = 0x80000000


# Shell Folder Get Attributes Options
SFGAOF = c_ulong


class SFGAO(IntFlag):
    SFGAO_CANCOPY = 0x00000001  # Objects can be copied.
    SFGAO_CANMOVE = 0x00000002  # Objects can be moved.
    SFGAO_CANLINK = 0x00000004  # Objects can be linked.
    SFGAO_STORAGE = 0x00000008  # Objects can be stored.
    SFGAO_CANRENAME = 0x00000010  # Objects can be renamed.
    SFGAO_CANDELETE = 0x00000020  # Objects can be deleted.
    SFGAO_HASPROPSHEET = 0x00000040  # Objects have property sheets.
    SFGAO_DROPTARGET = 0x00000100  # Objects are drop targets.
    SFGAO_CAPABILITYMASK = 0x00000177  # Mask for all capability flags.
    SFGAO_ENCRYPTED = 0x00002000  # Object is encrypted (use alt color).
    SFGAO_ISSLOW = 0x00004000  # Accessing this object is slow.
    SFGAO_GHOSTED = 0x00008000  # Object is ghosted (dimmed).
    SFGAO_LINK = 0x00010000  # Shortcut (link).
    SFGAO_SHARE = 0x00020000  # Shared.
    SFGAO_READONLY = 0x00040000  # Read-only.
    SFGAO_HIDDEN = 0x00080000  # Hidden object.
    SFGAO_DISPLAYATTRMASK = 0x000FC000  # Mask for display attributes.
    SFGAO_FILESYSANCESTOR = 0x10000000  # May contain children with file system folders.
    SFGAO_FOLDER = 0x20000000  # Is a folder.
    SFGAO_FILESYSTEM = 0x40000000  # Is part of the file system.
    SFGAO_HASSUBFOLDER = 0x80000000  # May contain subfolders.
    SFGAO_CONTENTSMASK = 0x80000000  # Mask for contents.
    SFGAO_VALIDATE = 0x01000000  # Invalidate cached information.
    SFGAO_REMOVABLE = 0x02000000  # Is a removable media.
    SFGAO_COMPRESSED = 0x04000000  # Object is compressed.
    SFGAO_BROWSABLE = 0x08000000  # Supports browsing.
    SFGAO_NONENUMERATED = 0x00100000  # Is not enumerated.
    SFGAO_NEWCONTENT = 0x00200000  # New content is present.
    SFGAO_CANMONIKER = 0x00400000  # Can create monikers for this item.
    SFGAO_HASSTORAGE = 0x00400000  # Supports storage interfaces.
    SFGAO_STREAM = 0x00400000  # Is a stream object.
    SFGAO_STORAGEANCESTOR = 0x00800000  # May contain children with storage folders.
    SFGAO_STORAGECAPMASK = 0x70C50008  # Mask for storage capability attributes.
    SFGAO_PKEYSFGAOMASK = 0x81044000  # Attributes that are part of the PKEY_SFGAOFlags property.


class SIGDN(c_int):
    SIGDN_NORMALDISPLAY = 0x00000000
    SIGDN_PARENTRELATIVEPARSING = 0x80018001
    SIGDN_PARENTRELATIVEFORADDRESSBAR = 0x8001C001
    SIGDN_DESKTOPABSOLUTEPARSING = 0x80028000
    SIGDN_PARENTRELATIVEEDITING = 0x80031001
    SIGDN_DESKTOPABSOLUTEEDITING = 0x8004C000
    SIGDN_FILESYSPATH = 0x80058000
    SIGDN_URL = 0x80068000


class FDAP(c_int):
    FDAP_BOTTOM = 0x00000000
    FDAP_TOP = 0x00000001


class FDE_SHAREVIOLATION_RESPONSE(c_int):  # noqa: N801
    FDESVR_DEFAULT = 0x00000000
    FDESVR_ACCEPT = 0x00000001
    FDESVR_REFUSE = 0x00000002


FDE_OVERWRITE_RESPONSE = FDE_SHAREVIOLATION_RESPONSE


class COMFunctionPointers:
    def __init__(self):
        self.hOle32: _FuncPointer
        self.hShell32: _FuncPointer
        self.pCoInitialize: _FuncPointer
        self.pCoUninitialize: _FuncPointer
        self.pCoCreateInstance: _FuncPointer
        self.pCoTaskMemFree: _FuncPointer
        self.pSHCreateItemFromParsingName: _FuncPointer

    @staticmethod
    def load_library(dll_name: str) -> _FuncPointer:
        windll.kernel32.LoadLibraryW.argtypes = [LPCWSTR]
        windll.kernel32.LoadLibraryW.restype = c_void_p
        handle = windll.kernel32.LoadLibraryW(dll_name)
        if not handle:
            raise ValueError(f"Unable to load library: {dll_name}")
        return handle

    @staticmethod
    def resolve_function(handle: _FuncPointer, func: bytes, func_type: type[_FuncPointer]) -> _FuncPointer:
        windll.kernel32.GetProcAddress.argtypes = [c_void_p, c_char_p]
        windll.kernel32.GetProcAddress.restype = c_void_p
        address = windll.kernel32.GetProcAddress(handle, func)
        assert address is not None
        return func_type(address)


class COMDLG_FILTERSPEC(Structure):  # noqa: N801
    _fields_: Sequence[tuple[str, type[_CData]] | tuple[str, type[_CData], int]] = [("pszName", LPCWSTR), ("pszSpec", LPCWSTR)]


class IModalWindow(comtypes.IUnknown):
    _case_insensitive_: bool = True
    _iid_: GUID = IID_IModalWindow
    _methods_: ClassVar[list[_ComMemberSpec]] = [COMMETHOD([], HRESULT, "Show", (["in"], HWND, "hwndParent"))]
    Show: Callable[[int | HWND], HRESULT]


class ModalWindow(comtypes.COMObject):
    _com_interfaces_: Sequence[type[comtypes.IUnknown]] = [IModalWindow]

    def QueryInterface(self, riid: GUID, ppvObject: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK

    def AddRef(self) -> ULONG:
        return ULONG(-1)

    def Release(self) -> ULONG:
        return ULONG(-1)

    def Show(self, hwndParent: HWND | int) -> HRESULT:  # noqa: N803
        return S_OK


class IShellItem(comtypes.IUnknown):
    _case_insensitive_: bool = True
    _iid_: GUID = IID_IShellItem
    _methods_: ClassVar[list[_ComMemberSpec]] = [
        COMMETHOD(
            [],
            HRESULT,
            "BindToHandler",
            (["in"], POINTER(comtypes.IUnknown), "pbc"),
            (["in"], POINTER(GUID), "bhid"),
            (["in"], POINTER(GUID), "riid"),
            (["out"], POINTER(c_void_p), "ppv"),
        ),
        COMMETHOD([], HRESULT, "GetParent", (["out"], POINTER(POINTER(comtypes.IUnknown)), "ppsi")),
        COMMETHOD([], HRESULT, "GetDisplayName", (["in"], c_ulong, "sigdnName"), (["out"], POINTER(LPWSTR), "ppszName")),
        COMMETHOD([], HRESULT, "GetAttributes", (["in"], c_ulong, "sfgaoMask"), (["out"], POINTER(c_ulong), "psfgaoAttribs")),
        COMMETHOD([], HRESULT, "Compare", (["in"], POINTER(comtypes.IUnknown), "psi"), (["in"], c_ulong, "hint"), (["out"], POINTER(c_int), "piOrder")),
    ]
    QueryInterface: Callable[[GUID, _Pointer[_Pointer[comtypes.IUnknown]]], HRESULT]
    AddRef: Callable[[], ULONG]
    Release: Callable[[], ULONG]
    BindToHandler: Callable[[_Pointer[comtypes.IUnknown], GUID, GUID, _Pointer[c_void_p]], HRESULT]
    GetParent: Callable[[], comtypes.IUnknown]
    GetDisplayName: Callable[[c_ulong | int], str]
    GetAttributes: Callable[[c_ulong | int], int]
    Compare: Callable[[_Pointer[comtypes.IUnknown], c_ulong, c_int], HRESULT]

    @classmethod
    def from_path(cls, path: os.PathLike | str) -> _Pointer[Self]:
        pShellItem = POINTER(cls)(cls())
        hr = SHCreateItemFromParsingName(str(path), None, byref(pShellItem))
        if hr != 0:
            raise OSError(f"Failed to create IShellItem from path. HRESULT: {hr}")
        return pShellItem


class IShellItem2(IShellItem):
    _iid_ = IID_IShellItem2
    _methods_: ClassVar[list[_ComMemberSpec]] = [
        COMMETHOD([], HRESULT, "GetPropertyStore", (["in"], c_uint, "flags"), (["in"], POINTER(GUID), "riid"), (["out"], POINTER(c_void_p), "ppv")),
        COMMETHOD(
            [],
            HRESULT,
            "GetPropertyStoreWithCreateObject",
            (["in"], c_uint, "flags"),
            (["in"], POINTER(comtypes.IUnknown), "punkCreateObject"),
            (["in"], POINTER(GUID), "riid"),
            (["out"], POINTER(c_void_p), "ppv"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "GetPropertyStoreForKeys",
            (["in"], POINTER(GUID), "rgKeys"),
            (["in"], c_uint, "cKeys"),
            (["in"], c_uint, "flags"),
            (["in"], POINTER(GUID), "riid"),
            (["out"], POINTER(c_void_p), "ppv"),
        ),
        COMMETHOD([], HRESULT, "GetPropertyDescriptionList", (["in"], POINTER(GUID), "keyType"), (["in"], POINTER(GUID), "riid"), (["out"], POINTER(c_void_p), "ppv")),
        COMMETHOD([], HRESULT, "Update", (["in"], POINTER(comtypes.IUnknown), "pbc")),
        COMMETHOD([], HRESULT, "GetProperty", (["in"], POINTER(GUID), "key"), (["out"], POINTER(c_void_p), "ppv")),
        COMMETHOD([], HRESULT, "GetCLSID", (["in"], POINTER(GUID), "key"), (["out"], POINTER(GUID), "pclsid")),
        COMMETHOD([], HRESULT, "GetFileTime", (["in"], POINTER(GUID), "key"), (["out"], POINTER(c_void_p), "pft")),
        COMMETHOD([], HRESULT, "GetInt32", (["in"], POINTER(GUID), "key"), (["out"], POINTER(c_int), "pi")),
        COMMETHOD([], HRESULT, "GetString", (["in"], POINTER(GUID), "key"), (["out"], POINTER(LPWSTR), "ppsz")),
        COMMETHOD([], HRESULT, "GetUInt32", (["in"], POINTER(GUID), "key"), (["out"], POINTER(c_uint), "pui")),
        COMMETHOD([], HRESULT, "GetUInt64", (["in"], POINTER(GUID), "key"), (["out"], POINTER(c_ulonglong), "pull")),
        COMMETHOD([], HRESULT, "GetBool", (["in"], POINTER(GUID), "key"), (["out"], POINTER(c_int), "pf")),
    ]


class ShellItem(comtypes.COMObject):
    _com_interfaces_: Sequence[type[comtypes.IUnknown]] = [IShellItem]

    def QueryInterface(self, riid: GUID, ppvObject: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK

    def AddRef(self) -> ULONG:
        return ULONG(-1)

    def Release(self) -> ULONG:
        return ULONG(-1)

    def BindToHandler(self, pbc: _Pointer[comtypes.IUnknown], bhid: GUID, riid: GUID, ppv: _Pointer[c_void_p]) -> HRESULT:
        return S_OK

    def GetParent(self, ppsi: comtypes.IUnknown) -> HRESULT:
        return S_OK

    def GetDisplayName(self, sigdnName: c_ulong | int, ppszName: _Pointer[c_wchar_p]) -> HRESULT:  # noqa: N803
        return S_OK

    def GetAttributes(self, sfgaoMask: c_ulong | int, psfgaoAttribs: _Pointer[c_ulong]) -> HRESULT:  # noqa: N803
        return S_OK

    def Compare(self, psi: _Pointer[comtypes.IUnknown], hint: c_ulong | int, piOrder: c_int) -> HRESULT:  # noqa: N803
        return S_OK


SHCreateItemFromParsingName = windll.shell32.SHCreateItemFromParsingName
SHCreateItemFromParsingName.argtypes = [LPCWSTR, comtypes.POINTER(comtypes.IUnknown), comtypes.POINTER(POINTER(IShellItem))]
SHCreateItemFromParsingName.restype = HRESULT


def create_shell_item_from_path(path: str) -> _Pointer[IShellItem]:
    item = POINTER(IShellItem)()
    hr = SHCreateItemFromParsingName(path, None, byref(GUID("{00000000-0000-0000-C000-000000000046}")), byref(item))
    if hr != 0:
        raise OSError(f"SHCreateItemFromParsingName failed! HRESULT: {hr}")
    return item


class IContextMenu(comtypes.IUnknown):
    _case_insensitive_: bool = True
    _iid_: GUID = IID_IContextMenu
    _methods_: ClassVar[list[_ComMemberSpec]] = [
        COMMETHOD(
            [],
            HRESULT,
            "QueryContextMenu",
            (["in"], c_void_p, "hmenu"),
            (["in"], c_uint, "indexMenu"),
            (["in"], c_uint, "idCmdFirst"),
            (["in"], c_uint, "idCmdLast"),
            (["in"], c_uint, "uFlags"),
        ),
        COMMETHOD([], HRESULT, "InvokeCommand", (["in"], c_void_p, "pici")),
        COMMETHOD(
            [],
            HRESULT,
            "GetCommandString",
            (["in"], c_uint, "idCmd"),
            (["in"], c_uint, "uType"),
            (["in"], c_void_p, "pReserved"),
            (["out"], c_wchar_p, "pszName"),
            (["in"], c_uint, "cchMax"),
        ),
    ]
    QueryInterface: Callable[[GUID, _Pointer[_Pointer[comtypes.IUnknown]]], HRESULT]
    AddRef: Callable[[], ULONG]
    Release: Callable[[], ULONG]
    QueryContextMenu: Callable[[c_void_p, c_uint, c_uint, c_uint, c_uint], HRESULT]
    InvokeCommand: Callable[[c_void_p], HRESULT]
    GetCommandString: Callable[[c_uint, c_uint, c_void_p, _Pointer[c_wchar_p], c_uint], HRESULT]


class ContextMenu(comtypes.COMObject):
    _com_interfaces_: Sequence[type[comtypes.IUnknown]] = [IContextMenu]

    def QueryInterface(self, riid: GUID, ppvObject: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK

    def AddRef(self) -> ULONG:
        return ULONG(-1)

    def Release(self) -> ULONG:
        return ULONG(-1)

    def QueryContextMenu(self, hmenu: c_void_p, indexMenu: c_uint | int, idCmdFirst: c_uint | int, idCmdLast: c_uint | int, uFlags: c_uint | int) -> HRESULT:  # noqa: N803
        return S_OK

    def InvokeCommand(self, pici: c_void_p) -> HRESULT:
        return S_OK

    def GetCommandString(self, idCmd: c_uint | int, uType: c_uint | int, pReserved: c_void_p, pszName: c_wchar_p, cchMax: c_uint | int) -> HRESULT:  # noqa: N803
        return S_OK


class IShellFolder(comtypes.IUnknown):
    _case_insensitive_: bool = True
    _iid_: GUID = IID_IShellFolder
    _methods_: ClassVar[list[_ComMemberSpec]] = [
        COMMETHOD(
            [],
            HRESULT,
            "ParseDisplayName",
            (["in"], HWND, "hwnd"),
            (["in"], POINTER(comtypes.IUnknown), "pbc"),
            (["in"], LPCWSTR, "pszDisplayName"),
            (["out"], POINTER(ULONG), "pchEaten"),
            (["out"], POINTER(c_void_p), "ppidl"),
            (["in"], POINTER(ULONG), "pdwAttributes"),
        ),
        COMMETHOD([], HRESULT, "EnumObjects", (["in"], HWND, "hwnd"), (["in"], c_ulong, "grfFlags"), (["out"], POINTER(POINTER(comtypes.IUnknown)), "ppenumIDList")),
        COMMETHOD(
            [],
            HRESULT,
            "BindToObject",
            (["in"], c_void_p, "pidl"),
            (["in"], POINTER(comtypes.IUnknown), "pbc"),
            (["in"], POINTER(GUID), "riid"),
            (["out"], POINTER(c_void_p), "ppv"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "BindToStorage",
            (["in"], c_void_p, "pidl"),
            (["in"], POINTER(comtypes.IUnknown), "pbc"),
            (["in"], POINTER(GUID), "riid"),
            (["out"], POINTER(c_void_p), "ppv"),
        ),
        COMMETHOD([], HRESULT, "CompareIDs", (["in"], c_void_p, "lParam"), (["in"], c_void_p, "pidl1"), (["in"], c_void_p, "pidl2")),
        COMMETHOD([], HRESULT, "CreateViewObject", (["in"], HWND, "hwndOwner"), (["in"], POINTER(GUID), "riid"), (["out"], POINTER(c_void_p), "ppv")),
        COMMETHOD([], HRESULT, "GetAttributesOf", (["in"], c_uint, "cidl"), (["in"], C_POINTER(c_void_p), "apidl"), (["out"], POINTER(c_ulong), "rgfInOut")),
        COMMETHOD(
            [],
            HRESULT,
            "GetUIObjectOf",
            (["in"], HWND, "hwndOwner"),
            (["in"], c_uint, "cidl"),
            (["in"], C_POINTER(c_void_p), "apidl"),
            (["in"], POINTER(GUID), "riid"),
            (["in"], POINTER(c_uint), "rgfReserved"),
            (["out"], POINTER(c_void_p), "ppv"),
        ),
        COMMETHOD([], HRESULT, "GetDisplayNameOf", (["in"], c_void_p, "pidl"), (["in"], c_ulong, "uFlags"), (["out"], POINTER(c_wchar_p), "pName")),
        COMMETHOD(
            [],
            HRESULT,
            "SetNameOf",
            (["in"], HWND, "hwnd"),
            (["in"], c_void_p, "pidl"),
            (["in"], LPCWSTR, "pszName"),
            (["in"], c_ulong, "uFlags"),
            (["out"], POINTER(c_void_p), "ppidlOut"),
        ),
    ]
    QueryInterface: Callable[[GUID, _Pointer[_Pointer[comtypes.IUnknown]]], HRESULT]
    AddRef: Callable[[], ULONG]
    Release: Callable[[], ULONG]
    ParseDisplayName: Callable[[HWND, _Pointer[comtypes.IUnknown], LPCWSTR, _Pointer[ULONG], _Pointer[c_void_p], _Pointer[ULONG]], HRESULT]
    EnumObjects: Callable[[HWND, c_ulong, _Pointer[_Pointer[comtypes.IUnknown]]], HRESULT]
    BindToObject: Callable[[c_void_p, _Pointer[comtypes.IUnknown], GUID, _Pointer[c_void_p]], HRESULT]
    BindToStorage: Callable[[c_void_p, _Pointer[comtypes.IUnknown], GUID, _Pointer[c_void_p]], HRESULT]
    CompareIDs: Callable[[c_void_p, c_void_p, c_void_p], HRESULT]
    CreateViewObject: Callable[[HWND, GUID, _Pointer[c_void_p]], HRESULT]
    GetAttributesOf: Callable[[c_uint, _Pointer[c_void_p], _Pointer[c_ulong]], HRESULT]
    GetUIObjectOf: Callable[[HWND, c_uint, _Pointer[c_void_p], GUID, _Pointer[c_uint], _Pointer[c_void_p]], HRESULT]
    GetDisplayNameOf: Callable[[c_void_p, c_ulong, _Pointer[c_wchar_p]], HRESULT]
    SetNameOf: Callable[[HWND, c_void_p, LPCWSTR, c_ulong, _Pointer[c_void_p]], HRESULT]


class ShellFolder(comtypes.COMObject):
    _com_interfaces_: Sequence[type[comtypes.IUnknown]] = [IShellFolder]

    def QueryInterface(self, riid: GUID, ppvObject: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK

    def AddRef(self) -> ULONG:
        return ULONG(-1)

    def Release(self) -> ULONG:
        return ULONG(-1)

    def ParseDisplayName(
        self,
        hwnd: HWND | int,
        pbc: _Pointer[comtypes.IUnknown],
        pszDisplayName: LPCWSTR | str,
        pchEaten: _Pointer[ULONG],
        ppidl: _Pointer[c_void_p],
        pdwAttributes: _Pointer[ULONG],
    ) -> HRESULT:  # noqa: N803, PLR0913
        return S_OK

    def EnumObjects(self, hwnd: HWND | int, grfFlags: c_ulong | int, ppenumIDList: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK

    def BindToObject(self, pidl: c_void_p, pbc: _Pointer[comtypes.IUnknown], riid: GUID, ppv: _Pointer[c_void_p]) -> HRESULT:
        return S_OK

    def BindToStorage(self, pidl: c_void_p, pbc: _Pointer[comtypes.IUnknown], riid: GUID, ppv: _Pointer[c_void_p]) -> HRESULT:
        return S_OK

    def CompareIDs(self, lParam: c_void_p, pidl1: c_void_p, pidl2: c_void_p) -> HRESULT:  # noqa: N803
        return S_OK

    def CreateViewObject(self, hwndOwner: HWND | int, riid: GUID, ppv: _Pointer[c_void_p]) -> HRESULT:  # noqa: N803
        return S_OK

    def GetAttributesOf(self, cidl: c_uint | int, apidl: _Pointer[c_void_p], rgfInOut: _Pointer[c_ulong]) -> HRESULT:  # noqa: N803
        return S_OK

    def GetUIObjectOf(
        self, hwndOwner: HWND | int, cidl: c_uint | int, apidl: _Pointer[c_void_p], riid: GUID, rgfReserved: _Pointer[c_uint], ppv: _Pointer[c_void_p]
    ) -> HRESULT:  # noqa: N803, PLR0913, E501
        return S_OK

    def GetDisplayNameOf(self, pidl: c_void_p, uFlags: c_ulong | int, pName: _Pointer[c_wchar_p]) -> HRESULT:  # noqa: N803
        return S_OK

    def SetNameOf(self, hwnd: HWND | int, pidl: c_void_p, pszName: LPCWSTR | str, uFlags: c_ulong | int, ppidlOut: _Pointer[c_void_p]) -> HRESULT:  # noqa: N803
        return S_OK


class IShellItemArray(comtypes.IUnknown):
    _case_insensitive_: bool = True
    _iid_: GUID = IID_IShellItemArray
    _methods_: ClassVar[list[_ComMemberSpec]] = [
        COMMETHOD(
            [],
            HRESULT,
            "BindToHandler",
            (["in"], POINTER(comtypes.IUnknown), "pbc"),
            (["in"], POINTER(GUID), "bhid"),
            (["in"], POINTER(GUID), "riid"),
            (["out"], POINTER(c_void_p), "ppv"),
        ),
        COMMETHOD([], HRESULT, "GetPropertyStore", (["in"], c_ulong, "flags"), (["in"], POINTER(GUID), "riid"), (["out"], POINTER(c_void_p), "ppv")),
        COMMETHOD([], HRESULT, "GetPropertyDescriptionList", (["in"], POINTER(GUID), "keyType"), (["in"], POINTER(GUID), "riid"), (["out"], POINTER(c_void_p), "ppv")),
        COMMETHOD([], HRESULT, "GetAttributes", (["in"], c_ulong, "attribFlags"), (["in"], c_ulong, "sfgaoMask"), (["out"], POINTER(c_ulong), "psfgaoAttribs")),
        COMMETHOD([], HRESULT, "GetCount", (["out"], POINTER(c_uint), "pdwNumItems")),
        COMMETHOD([], HRESULT, "GetItemAt", (["in"], c_uint, "dwIndex"), (["out"], POINTER(POINTER(IShellItem)), "ppsi")),
        COMMETHOD([], HRESULT, "EnumItems", (["out"], POINTER(POINTER(comtypes.IUnknown)), "ppenumShellItems")),
    ]
    QueryInterface: Callable[[GUID, comtypes.IUnknown], HRESULT]
    AddRef: Callable[[], ULONG]
    Release: Callable[[], ULONG]
    BindToHandler: Callable[[_Pointer[comtypes.IUnknown], GUID, GUID], int]
    GetPropertyStore: Callable[[c_ulong, GUID], c_void_p]
    GetPropertyDescriptionList: Callable[[GUID, GUID], c_void_p]
    GetAttributes: Callable[[c_ulong, c_ulong], _Pointer[c_ulong]]
    GetCount: Callable[[], int]
    GetItemAt: Callable[[c_uint | int], IShellItem]
    EnumItems: Callable[[], comtypes.IUnknown]


class ShellItemArray(comtypes.COMObject):
    _com_interfaces_: Sequence[type[comtypes.IUnknown]] = [IShellItemArray]

    def QueryInterface(self, riid: GUID, ppvObject: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK

    def AddRef(self) -> ULONG:
        return ULONG(-1)

    def Release(self) -> ULONG:
        return ULONG(-1)

    def BindToHandler(self, pbc: _Pointer[comtypes.IUnknown], bhid: GUID, riid: GUID, ppv: _Pointer[c_void_p]) -> HRESULT:
        return S_OK

    def GetPropertyStore(self, flags: c_ulong | int, riid: GUID, ppv: _Pointer[c_void_p]) -> HRESULT:
        return S_OK

    def GetPropertyDescriptionList(self, keyType: GUID, riid: GUID, ppv: _Pointer[c_void_p]) -> HRESULT:  # noqa: N803
        return S_OK

    def GetAttributes(self, attribFlags: c_ulong | int, sfgaoMask: c_ulong | int, psfgaoAttribs: _Pointer[c_ulong]) -> HRESULT:  # noqa: N803
        return S_OK

    def GetCount(self, pdwNumItems: _Pointer[c_uint]) -> HRESULT:  # noqa: N803
        return S_OK

    def GetItemAt(self, dwIndex: c_uint | int, ppsi: IShellItem) -> HRESULT:  # noqa: N803
        return S_OK

    def EnumItems(self, ppenumShellItems: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK


class IShellItemFilter(comtypes.IUnknown):
    _case_insensitive_: bool = True
    _iid_: GUID = IID_IShellItemFilter
    _methods_: ClassVar[list[_ComMemberSpec]] = [
        COMMETHOD([], HRESULT, "IncludeItem", (["in"], POINTER(IShellItem), "psi")),
        COMMETHOD([], HRESULT, "GetEnumFlagsForItem", (["in"], POINTER(IShellItem), "psi"), (["out"], POINTER(c_ulong), "pgrfFlags")),
    ]
    QueryInterface: Callable[[GUID, comtypes.IUnknown], HRESULT]
    AddRef: Callable[[], ULONG]
    Release: Callable[[], ULONG]
    IncludeItem: Callable[[IShellItem], c_ulong]
    GetEnumFlagsForItem: Callable[[], HRESULT]


class ShellItemFilter(comtypes.COMObject):
    _com_interfaces_: Sequence[type[comtypes.IUnknown]] = [IShellItemFilter]

    def QueryInterface(self, riid: GUID, ppvObject: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK

    def AddRef(self) -> ULONG:
        return ULONG(-1)

    def Release(self) -> ULONG:
        return ULONG(-1)

    def IncludeItem(self, psi: IShellItem) -> HRESULT:
        return S_OK

    def GetEnumFlagsForItem(self, psi: IShellItem, pgrfFlags: _Pointer[c_ulong]) -> HRESULT:  # noqa: N803
        return S_OK


class IEnumShellItems(comtypes.IUnknown):
    _case_insensitive_: bool = True
    _iid_: GUID = IID_IEnumShellItems
    _methods_: ClassVar[list[_ComMemberSpec]]
    QueryInterface: Callable[[GUID, comtypes.IUnknown], HRESULT]
    AddRef: Callable[[], ULONG]
    Release: Callable[[], ULONG]
    Next: Callable[[_Pointer[IEnumShellItems], c_ulong, IShellItem, _Pointer[c_ulong]], HRESULT]
    Skip: Callable[[_Pointer[IEnumShellItems], c_ulong], HRESULT]
    Reset: Callable[[_Pointer[IEnumShellItems]], HRESULT]
    Clone: Callable[[_Pointer[IEnumShellItems], _Pointer[_Pointer[IEnumShellItems]]], HRESULT]


IEnumShellItems._methods_ = [  # noqa: SLF001
    COMMETHOD([], HRESULT, "Next", (["in"], c_ulong, "celt"), (["out"], POINTER(POINTER(IShellItem)), "rgelt"), (["out"], POINTER(c_ulong), "pceltFetched")),
    COMMETHOD([], HRESULT, "Skip", (["in"], c_ulong, "celt")),
    COMMETHOD([], HRESULT, "Reset"),
    COMMETHOD([], HRESULT, "Clone", (["out"], POINTER(POINTER(IEnumShellItems)), "ppenum")),
]


class EnumShellItems(comtypes.COMObject):
    _com_interfaces_: Sequence[type[comtypes.IUnknown]] = [IEnumShellItems]

    def QueryInterface(self, riid: GUID, ppvObject: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK

    def AddRef(self) -> ULONG:
        return ULONG(-1)

    def Release(self) -> ULONG:
        return ULONG(-1)

    def Next(self, celt: c_ulong | int, rgelt: IShellItem, pceltFetched: _Pointer[c_ulong]) -> HRESULT:  # noqa: N803
        return S_OK

    def Skip(self, celt: c_ulong | int) -> HRESULT:
        return S_OK

    def Reset(self) -> HRESULT:
        return S_OK

    def Clone(self, ppenum: _Pointer[_Pointer[IEnumShellItems]]) -> HRESULT:
        return S_OK


class IPropertyStore(comtypes.IUnknown):
    _case_insensitive_: bool = True
    _iid_: GUID = IID_IPropertyStore
    _methods_: ClassVar[list[_ComMemberSpec]] = [
        COMMETHOD([], HRESULT, "GetCount", (["out"], POINTER(c_ulong), "count")),
        COMMETHOD([], HRESULT, "GetAt", (["in"], c_ulong, "index"), (["out"], POINTER(GUID), "key")),
        COMMETHOD([], HRESULT, "GetValue", (["in"], POINTER(GUID), "key"), (["out"], POINTER(c_void_p), "pv")),
        COMMETHOD([], HRESULT, "SetValue", (["in"], POINTER(GUID), "key"), (["in"], POINTER(c_void_p), "propvar")),
        COMMETHOD([], HRESULT, "Commit"),
    ]
    QueryInterface: Callable[[GUID, comtypes.IUnknown], HRESULT]
    AddRef: Callable[[], ULONG]
    Release: Callable[[], ULONG]
    GetCount: Callable[[_Pointer[IPropertyStore], _Pointer[c_ulong]], HRESULT]
    GetAt: Callable[[_Pointer[IPropertyStore], c_ulong, GUID], HRESULT]
    GetValue: Callable[[_Pointer[IPropertyStore], GUID, _Pointer[c_void_p]], HRESULT]
    SetValue: Callable[[_Pointer[IPropertyStore], GUID, _Pointer[c_void_p]], HRESULT]
    Commit: Callable[[_Pointer[IPropertyStore]], HRESULT]


class PropertyStore(comtypes.COMObject):
    _com_interfaces_: Sequence[type[comtypes.IUnknown]] = [IPropertyStore]

    def QueryInterface(self, riid: GUID, ppvObject: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK

    def AddRef(self) -> ULONG:
        return ULONG(-1)

    def Release(self) -> ULONG:
        return ULONG(-1)

    def GetCount(self, count: _Pointer[c_ulong]) -> HRESULT:
        return S_OK

    def GetAt(self, index: c_ulong | int, key: GUID) -> HRESULT:
        return S_OK

    def GetValue(self, key: GUID, pv: _Pointer[c_void_p]) -> HRESULT:
        return S_OK

    def SetValue(self, key: GUID, propvar: _Pointer[c_void_p]) -> HRESULT:
        return S_OK

    def Commit(self) -> HRESULT:
        return S_OK


class IFileOperationProgressSink(comtypes.IUnknown):
    _case_insensitive_: bool = True
    _iid_: GUID = IID_IFileOperationProgressSink
    _methods_: ClassVar[list[_ComMemberSpec]] = [
        COMMETHOD([], HRESULT, "StartOperations"),
        COMMETHOD([], HRESULT, "FinishOperations", (["in"], HRESULT, "hr")),
        COMMETHOD([], HRESULT, "PreRenameItem", (["in"], c_ulong, "dwFlags"), (["in"], POINTER(IShellItem), "psiItem"), (["in"], c_wchar_p, "pszNewName")),
        COMMETHOD(
            [],
            HRESULT,
            "PostRenameItem",
            (["in"], c_ulong, "dwFlags"),
            (["in"], POINTER(IShellItem), "psiItem"),
            (["in"], c_wchar_p, "pszNewName"),
            (["in"], HRESULT, "hrRename"),
            (["in"], POINTER(IShellItem), "psiNewlyCreated"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "PreMoveItem",
            (["in"], c_ulong, "dwFlags"),
            (["in"], POINTER(IShellItem), "psiItem"),
            (["in"], POINTER(IShellItem), "psiDestinationFolder"),
            (["in"], c_wchar_p, "pszNewName"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "PostMoveItem",
            (["in"], c_ulong, "dwFlags"),
            (["in"], POINTER(IShellItem), "psiItem"),
            (["in"], POINTER(IShellItem), "psiDestinationFolder"),
            (["in"], c_wchar_p, "pszNewName"),
            (["in"], HRESULT, "hrMove"),
            (["in"], POINTER(IShellItem), "psiNewlyCreated"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "PreCopyItem",
            (["in"], c_ulong, "dwFlags"),
            (["in"], POINTER(IShellItem), "psiItem"),
            (["in"], POINTER(IShellItem), "psiDestinationFolder"),
            (["in"], c_wchar_p, "pszNewName"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "PostCopyItem",
            (["in"], c_ulong, "dwFlags"),
            (["in"], POINTER(IShellItem), "psiItem"),
            (["in"], POINTER(IShellItem), "psiDestinationFolder"),
            (["in"], c_wchar_p, "pszNewName"),
            (["in"], HRESULT, "hrCopy"),
            (["in"], POINTER(IShellItem), "psiNewlyCreated"),
        ),
        COMMETHOD([], HRESULT, "PreDeleteItem", (["in"], c_ulong, "dwFlags"), (["in"], POINTER(IShellItem), "psiItem")),
        COMMETHOD(
            [],
            HRESULT,
            "PostDeleteItem",
            (["in"], c_ulong, "dwFlags"),
            (["in"], POINTER(IShellItem), "psiItem"),
            (["in"], HRESULT, "hrDelete"),
            (["in"], POINTER(IShellItem), "psiNewlyCreated"),
        ),
        COMMETHOD([], HRESULT, "PreNewItem", (["in"], c_ulong, "dwFlags"), (["in"], POINTER(IShellItem), "psiDestinationFolder"), (["in"], c_wchar_p, "pszNewName")),
        COMMETHOD(
            [],
            HRESULT,
            "PostNewItem",
            (["in"], c_ulong, "dwFlags"),
            (["in"], POINTER(IShellItem), "psiDestinationFolder"),
            (["in"], c_wchar_p, "pszNewName"),
            (["in"], c_wchar_p, "pszTemplateName"),
            (["in"], c_ulong, "dwFileAttributes"),
            (["in"], HRESULT, "hrNew"),
            (["in"], POINTER(IShellItem), "psiNewItem"),
        ),
        COMMETHOD([], HRESULT, "UpdateProgress", (["in"], c_ulong, "iWorkTotal"), (["in"], c_ulong, "iWorkSoFar")),
        COMMETHOD([], HRESULT, "ResetTimer"),
        COMMETHOD([], HRESULT, "PauseTimer"),
        COMMETHOD([], HRESULT, "ResumeTimer"),
    ]
    QueryInterface: Callable[[GUID, comtypes.IUnknown], HRESULT]
    AddRef: Callable[[], ULONG]
    Release: Callable[[], ULONG]
    StartOperations: Callable[[], HRESULT]
    FinishOperations: Callable[[HRESULT], HRESULT]
    PreRenameItem: Callable[[c_ulong, IShellItem, c_wchar_p], HRESULT]
    PostRenameItem: Callable[[c_ulong, IShellItem, c_wchar_p, HRESULT, IShellItem], HRESULT]
    PreMoveItem: Callable[[c_ulong, IShellItem, IShellItem, c_wchar_p], HRESULT]
    PostMoveItem: Callable[[c_ulong, IShellItem, IShellItem, c_wchar_p, HRESULT, IShellItem], HRESULT]
    PreCopyItem: Callable[[c_ulong, IShellItem, IShellItem, c_wchar_p], HRESULT]
    PostCopyItem: Callable[[c_ulong, IShellItem, IShellItem, c_wchar_p, HRESULT, IShellItem], HRESULT]
    PreDeleteItem: Callable[[c_ulong, IShellItem], HRESULT]
    PostDeleteItem: Callable[[c_ulong, IShellItem, HRESULT, IShellItem], HRESULT]
    PreNewItem: Callable[[c_ulong, IShellItem, c_wchar_p], HRESULT]
    PostNewItem: Callable[[c_ulong, IShellItem, c_wchar_p, c_wchar_p, c_ulong, HRESULT, IShellItem], HRESULT]
    UpdateProgress: Callable[[c_ulong, c_ulong], HRESULT]
    ResetTimer: Callable[[], HRESULT]
    PauseTimer: Callable[[], HRESULT]
    ResumeTimer: Callable[[], HRESULT]


class FileOperationProgressSink(comtypes.COMObject):
    _com_interfaces_: Sequence[type[comtypes.IUnknown]] = [IFileOperationProgressSink]

    def StartOperations(self) -> HRESULT:
        return S_OK

    def FinishOperations(self, hr: HRESULT) -> HRESULT:
        return S_OK

    def PreRenameItem(self, dwFlags: c_ulong | int, psiItem: IShellItem, pszNewName: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def PostRenameItem(self, dwFlags: c_ulong | int, psiItem: IShellItem, pszNewName: LPCWSTR | str, hrRename: HRESULT, psiNewlyCreated: IShellItem) -> HRESULT:  # noqa: N803
        return S_OK

    def PreMoveItem(self, dwFlags: c_ulong | int, psiItem: IShellItem, psiDestinationFolder: IShellItem, pszNewName: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def PostMoveItem(
        self, dwFlags: c_ulong | int, psiItem: IShellItem, psiDestinationFolder: IShellItem, pszNewName: LPCWSTR | str, hrMove: HRESULT, psiNewlyCreated: IShellItem
    ) -> HRESULT:  # noqa: N803, E501
        return S_OK

    def PreCopyItem(self, dwFlags: c_ulong | int, psiItem: IShellItem, psiDestinationFolder: IShellItem, pszNewName: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def PostCopyItem(
        self, dwFlags: c_ulong | int, psiItem: IShellItem, psiDestinationFolder: IShellItem, pszNewName: LPCWSTR | str, hrCopy: HRESULT, psiNewlyCreated: IShellItem
    ) -> HRESULT:  # noqa: N803, E501
        return S_OK

    def PreDeleteItem(self, dwFlags: c_ulong | int, psiItem: IShellItem) -> HRESULT:
        return S_OK

    def PostDeleteItem(self, dwFlags: c_ulong | int, psiItem: IShellItem, hrDelete: HRESULT, psiNewlyCreated: IShellItem) -> HRESULT:
        return S_OK

    def PreNewItem(self, dwFlags: c_ulong | int, psiDestinationFolder: IShellItem, pszNewName: LPCWSTR | str) -> HRESULT:
        return S_OK

    def PostNewItem(
        self,
        dwFlags: c_ulong | int,
        psiDestinationFolder: IShellItem,
        pszNewName: LPCWSTR | str,
        pszTemplateName: LPCWSTR | str,
        dwFileAttributes: c_ulong | int,
        hrNew: HRESULT,
        psiNewItem: IShellItem,
    ) -> HRESULT:  # noqa: N803, E501
        return S_OK

    def UpdateProgress(self, iWorkTotal: c_ulong | int, iWorkSoFar: c_ulong | int) -> HRESULT:  # noqa: N803
        return S_OK

    def ResetTimer(self) -> HRESULT:
        return S_OK

    def PauseTimer(self) -> HRESULT:
        return S_OK

    def ResumeTimer(self) -> HRESULT:
        return S_OK


class IFileOperation(comtypes.IUnknown):
    _case_insensitive_ = True
    _iid_ = IID_IFileOperation
    _idlflags_ = []

    _methods_: ClassVar[list[_ComMemberSpec]] = [
        # Advise methods
        comtypes.COMMETHOD([], HRESULT, "Advise", (["in"], POINTER(IFileOperationProgressSink), "pfops"), (["out"], POINTER(c_ulong), "pdwCookie")),
        comtypes.COMMETHOD([], HRESULT, "Unadvise", (["in"], c_ulong, "dwCookie")),
        # Operation control methods
        comtypes.COMMETHOD([], HRESULT, "SetOperationFlags", (["in"], c_ulong, "dwOperationFlags")),
        comtypes.COMMETHOD([], HRESULT, "SetProgressMessage", (["in"], c_wchar_p, "pszMessage")),
        comtypes.COMMETHOD([], HRESULT, "SetProgressDialog", (["in"], POINTER(comtypes.IUnknown), "popd")),
        # Item methods
        comtypes.COMMETHOD([], HRESULT, "SetProperties", (["in"], POINTER(comtypes.IUnknown), "pproparray")),
        comtypes.COMMETHOD([], HRESULT, "SetOwnerWindow", (["in"], c_ulong, "hwndOwner")),
        comtypes.COMMETHOD([], HRESULT, "ApplyPropertiesToItem", (["in"], POINTER(IShellItem), "psiItem")),
        comtypes.COMMETHOD([], HRESULT, "ApplyPropertiesToItems", (["in"], POINTER(comtypes.IUnknown), "punkItems")),
        # Operation methods
        comtypes.COMMETHOD(
            [], HRESULT, "RenameItem", (["in"], POINTER(IShellItem), "psiItem"), (["in"], c_wchar_p, "pszNewName"), (["in"], POINTER(IFileOperationProgressSink), "pfopsItem")
        ),
        comtypes.COMMETHOD([], HRESULT, "RenameItems", (["in"], POINTER(comtypes.IUnknown), "pUnkItems"), (["in"], c_wchar_p, "pszNewName")),
        comtypes.COMMETHOD(
            [],
            HRESULT,
            "MoveItem",
            (["in"], POINTER(IShellItem), "psiItem"),
            (["in"], POINTER(IShellItem), "psiDestinationFolder"),
            (["in"], c_wchar_p, "pszNewName"),
            (["in"], POINTER(IFileOperationProgressSink), "pfopsItem"),
        ),
        comtypes.COMMETHOD([], HRESULT, "MoveItems", (["in"], POINTER(comtypes.IUnknown), "punkItems"), (["in"], POINTER(IShellItem), "psiDestinationFolder")),
        comtypes.COMMETHOD(
            [],
            HRESULT,
            "CopyItem",
            (["in"], POINTER(IShellItem), "psiItem"),
            (["in"], POINTER(IShellItem), "psiDestinationFolder"),
            (["in"], c_wchar_p, "pszNewName"),
            (["in"], POINTER(IFileOperationProgressSink), "pfopsItem"),
        ),
        comtypes.COMMETHOD([], HRESULT, "CopyItems", (["in"], POINTER(comtypes.IUnknown), "punkItems"), (["in"], POINTER(IShellItem), "psiDestinationFolder")),
        comtypes.COMMETHOD([], HRESULT, "DeleteItem", (["in"], POINTER(IShellItem), "psiItem"), (["in"], POINTER(IFileOperationProgressSink), "pfopsItem")),
        comtypes.COMMETHOD([], HRESULT, "DeleteItems", (["in"], POINTER(comtypes.IUnknown), "punkItems")),
        comtypes.COMMETHOD(
            [],
            HRESULT,
            "NewItem",
            (["in"], POINTER(IShellItem), "psiDestinationFolder"),
            (["in"], c_ulong, "dwFileAttributes"),
            (["in"], c_wchar_p, "pszName"),
            (["in"], c_wchar_p, "pszTemplateName"),
            (["in"], POINTER(IFileOperationProgressSink), "pfopsItem"),
        ),
        # Execution methods
        comtypes.COMMETHOD([], HRESULT, "PerformOperations"),
        comtypes.COMMETHOD([], HRESULT, "GetAnyOperationsAborted", (["out"], POINTER(c_int), "pfAnyOperationsAborted")),
    ]


class IFileDialogEvents(comtypes.IUnknown):
    _case_insensitive_: bool = True
    _iid_: GUID = IID_IFileDialogEvents
    _methods_: ClassVar[list[_ComMemberSpec]]
    QueryInterface: Callable[[GUID, comtypes.IUnknown], HRESULT]
    AddRef: Callable[[], ULONG]
    Release: Callable[[], ULONG]
    OnFileOk: Callable[[IFileDialog], HRESULT]
    OnFolderChanging: Callable[[IFileDialog, IShellItem], HRESULT]
    OnFolderChange: Callable[[IFileDialog], HRESULT]
    OnSelectionChange: Callable[[IFileDialog], HRESULT]
    OnShareViolation: Callable[[IFileDialog, IShellItem, c_int], HRESULT]
    OnTypeChange: Callable[[IFileDialog], HRESULT]
    OnOverwrite: Callable[[IFileDialog, IShellItem, c_int], HRESULT]


class IFileDialog(IModalWindow):
    _iid_: GUID = IID_IFileDialog
    _methods_: ClassVar[list[_ComMemberSpec]] = [
        COMMETHOD([], HRESULT, "SetFileTypes", (["in"], c_uint, "cFileTypes"), (["in"], POINTER(c_void_p), "rgFilterSpec")),
        COMMETHOD([], HRESULT, "SetFileTypeIndex", (["in"], c_uint, "iFileType")),
        COMMETHOD([], HRESULT, "GetFileTypeIndex", (["out"], POINTER(c_uint), "piFileType")),
        COMMETHOD([], HRESULT, "Advise", (["in"], POINTER(comtypes.IUnknown), "pfde"), (["out"], POINTER(DWORD), "pdwCookie")),
        COMMETHOD([], HRESULT, "Unadvise", (["in"], DWORD, "dwCookie")),
        COMMETHOD([], HRESULT, "SetOptions", (["in"], c_uint, "fos")),
        COMMETHOD([], HRESULT, "GetOptions", (["out"], POINTER(DWORD), "pfos")),
        COMMETHOD([], HRESULT, "SetDefaultFolder", (["in"], POINTER(IShellItem), "psi")),
        COMMETHOD([], HRESULT, "SetFolder", (["in"], POINTER(IShellItem), "psi")),
        COMMETHOD([], HRESULT, "GetFolder", (["out"], POINTER(POINTER(IShellItem)), "ppsi")),
        COMMETHOD([], HRESULT, "GetCurrentSelection", (["out"], POINTER(POINTER(IShellItem)), "ppsi")),
        COMMETHOD([], HRESULT, "SetFileName", (["in"], LPCWSTR, "pszName")),
        COMMETHOD([], HRESULT, "GetFileName", (["out"], POINTER(LPWSTR), "pszName")),
        COMMETHOD([], HRESULT, "SetTitle", (["in"], LPCWSTR, "pszTitle")),
        COMMETHOD([], HRESULT, "SetOkButtonLabel", (["in"], LPCWSTR, "pszText")),
        COMMETHOD([], HRESULT, "SetFileNameLabel", (["in"], LPCWSTR, "pszLabel")),
        COMMETHOD([], HRESULT, "GetResult", (["out"], POINTER(POINTER(IShellItem)), "ppsi")),
        COMMETHOD([], HRESULT, "AddPlace", (["in"], POINTER(IShellItem), "psi"), (["in"], c_int, "fdap")),
        COMMETHOD([], HRESULT, "SetDefaultExtension", (["in"], LPCWSTR, "pszDefaultExtension")),
        COMMETHOD([], HRESULT, "Close", (["in"], HRESULT, "hr")),
        COMMETHOD([], HRESULT, "SetClientGuid", (["in"], POINTER(GUID), "guid")),
        COMMETHOD([], HRESULT, "ClearClientData"),
        COMMETHOD([], HRESULT, "SetFilter", (["in"], POINTER(IShellItemFilter), "pFilter")),
    ]
    SetFileTypes: Callable[[c_uint | int, _Pointer[c_void_p]], HRESULT]
    SetFileTypeIndex: Callable[[c_uint], HRESULT]
    GetFileTypeIndex: Callable[[], _Pointer[c_uint]]
    Advise: Callable[[comtypes.IUnknown | comtypes.COMObject], int]
    Unadvise: Callable[[int], HRESULT]
    SetOptions: Callable[[DWORD | int], HRESULT]
    GetOptions: Callable[[], int]
    SetDefaultFolder: Callable[[_Pointer[IShellItem]], HRESULT]
    SetFolder: Callable[[_Pointer[IShellItem]], HRESULT]
    GetFolder: Callable[[], IShellItem]
    GetCurrentSelection: Callable[[], IShellItem]
    SetFileName: Callable[[str], HRESULT]
    GetFileName: Callable[[], _Pointer[LPWSTR]]
    SetTitle: Callable[[str], HRESULT]
    SetOkButtonLabel: Callable[[str], HRESULT]
    SetFileNameLabel: Callable[[str], HRESULT]
    GetResult: Callable[[], IShellItem]
    AddPlace: Callable[[IShellItem, c_int], HRESULT]
    SetDefaultExtension: Callable[[str], HRESULT]
    Close: Callable[[HRESULT], HRESULT]
    SetClientGuid: Callable[[GUID], HRESULT]
    ClearClientData: Callable[[], HRESULT]
    SetFilter: Callable[[IShellItemFilter], HRESULT]


IFileDialogEvents._methods_ = [  # noqa: SLF001
    COMMETHOD([], HRESULT, "OnFileOk", (["in"], POINTER(IFileDialog), "pfd")),
    COMMETHOD([], HRESULT, "OnFolderChanging", (["in"], POINTER(IFileDialog), "pfd"), (["in"], POINTER(IShellItem), "psiFolder")),
    COMMETHOD([], HRESULT, "OnFolderChange", (["in"], POINTER(IFileDialog), "pfd")),
    COMMETHOD([], HRESULT, "OnSelectionChange", (["in"], POINTER(IFileDialog), "pfd")),
    COMMETHOD([], HRESULT, "OnShareViolation", (["in"], POINTER(IFileDialog), "pfd"), (["in"], POINTER(IShellItem), "psi"), (["out"], POINTER(c_int), "pResponse")),
    COMMETHOD([], HRESULT, "OnTypeChange", (["in"], POINTER(IFileDialog), "pfd")),
    COMMETHOD([], HRESULT, "OnOverwrite", (["in"], POINTER(IFileDialog), "pfd"), (["in"], POINTER(IShellItem), "psi"), (["out"], POINTER(c_int), "pResponse")),
]


class FileDialogEvents(comtypes.COMObject):
    _com_interfaces_: Sequence[type[comtypes.IUnknown]] = [IFileDialogEvents]

    def QueryInterface(self, riid: GUID, ppvObject: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK

    def AddRef(self) -> ULONG:
        return ULONG(-1)

    def Release(self) -> ULONG:
        return ULONG(-1)

    def OnFileOk(self, ifd: IFileDialog) -> HRESULT:
        return S_OK

    def OnFolderChanging(self, ifd: IFileDialog, isiFolder: IShellItem) -> HRESULT:  # noqa: N803
        return S_OK

    def OnFolderChange(self, ifd: IFileDialog) -> HRESULT:
        return S_OK

    def OnSelectionChange(self, ifd: IFileDialog) -> HRESULT:
        return S_OK

    def OnShareViolation(self, ifd: IFileDialog, psi: IShellItem, response: c_int) -> HRESULT:
        return S_OK

    def OnTypeChange(self, ifd: IFileDialog) -> HRESULT:
        return S_OK

    def OnOverwrite(self, ifd: IFileDialog, psi: IShellItem, response: c_int) -> HRESULT:
        return S_OK


class FileDialog(comtypes.COMObject):
    _com_interfaces_: Sequence[type[comtypes.IUnknown]] = [IFileDialog]

    def QueryInterface(self, riid: GUID, ppvObject: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK

    def AddRef(self) -> ULONG:
        return ULONG(-1)

    def Release(self) -> ULONG:
        return ULONG(-1)

    def Show(self, hwndOwner: HWND | int) -> HRESULT:  # noqa: N803
        return S_OK

    def SetFileTypes(self, cFileTypes: c_uint | int, rgFilterSpec: Array[COMDLG_FILTERSPEC]) -> HRESULT:  # noqa: N803
        return S_OK

    def SetFileTypeIndex(self, iFileType: c_uint | int) -> HRESULT:  # noqa: N803
        return S_OK

    def GetFileTypeIndex(self, piFileType: _Pointer[c_uint]) -> HRESULT:  # noqa: N803
        return S_OK

    def Advise(self, pfde: _Pointer[comtypes.IUnknown], pdwCookie: _Pointer[DWORD]) -> HRESULT:  # noqa: N803
        return S_OK

    def Unadvise(self, dwCookie: int) -> HRESULT:  # noqa: N803
        return S_OK

    def SetOptions(self, fos: int) -> HRESULT:
        return S_OK

    def GetOptions(self, pfos: _Pointer[DWORD]) -> HRESULT:
        return S_OK

    def SetDefaultFolder(self, psi: IShellItem) -> HRESULT:
        return S_OK

    def SetFolder(self, psi: IShellItem) -> HRESULT:
        return S_OK

    def GetFolder(self, ppsi: IShellItem) -> HRESULT:
        return S_OK

    def GetCurrentSelection(self, ppsi: IShellItem) -> HRESULT:
        return S_OK

    def SetFileName(self, pszName: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def GetFileName(self, pszName: _Pointer[LPWSTR]) -> HRESULT:  # noqa: N803
        return S_OK

    def SetTitle(self, pszTitle: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def SetOkButtonLabel(self, pszText: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def SetFileNameLabel(self, pszLabel: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def GetResult(self, ppsi: IShellItem) -> HRESULT:
        return S_OK

    def AddPlace(self, psi: IShellItem, fdap: c_int) -> HRESULT:
        return S_OK

    def SetDefaultExtension(self, pszDefaultExtension: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def Close(self, hr: HRESULT | int) -> HRESULT:
        return S_OK

    def SetClientGuid(self, guid: GUID) -> HRESULT:
        return S_OK

    def ClearClientData(self) -> HRESULT:
        return S_OK

    def SetFilter(self, pFilter: IShellItemFilter) -> HRESULT:  # noqa: N803
        return S_OK


class IShellLibrary(comtypes.IUnknown):
    _case_insensitive_: bool = True
    _iid_: GUID = IID_IShellLibrary
    _methods_: ClassVar[list[_ComMemberSpec]] = [
        COMMETHOD([], HRESULT, "LoadLibraryFromItem", (["in"], POINTER(IShellItem), "psi"), (["in"], c_ulong, "grfMode")),
        COMMETHOD([], HRESULT, "LoadLibraryFromKnownFolder", (["in"], POINTER(GUID), "kfidLibrary"), (["in"], c_ulong, "grfMode")),
        COMMETHOD([], HRESULT, "AddFolder", (["in"], POINTER(IShellItem), "psi")),
        COMMETHOD([], HRESULT, "RemoveFolder", (["in"], POINTER(IShellItem), "psi")),
        COMMETHOD([], HRESULT, "GetFolders", (["in"], c_int, "lff"), (["in"], POINTER(GUID), "riid"), (["out"], POINTER(POINTER(c_void_p)), "ppv")),
        COMMETHOD(
            [],
            HRESULT,
            "ResolveFolder",
            (["in"], POINTER(IShellItem), "psi"),
            (["in"], c_ulong, "grfMode"),
            (["in"], POINTER(GUID), "riid"),
            (["out"], POINTER(POINTER(c_void_p)), "ppv"),
        ),
        COMMETHOD([], HRESULT, "GetDefaultSaveFolder", (["in"], c_int, "dsft"), (["in"], POINTER(GUID), "riid"), (["out"], POINTER(POINTER(c_void_p)), "ppv")),
        COMMETHOD([], HRESULT, "SetDefaultSaveFolder", (["in"], c_int, "dsft"), (["in"], POINTER(IShellItem), "psi")),
        COMMETHOD([], HRESULT, "GetOptions", (["out"], POINTER(c_uint), "pOptions")),
        COMMETHOD([], HRESULT, "SetOptions", (["in"], c_ulong, "stfOptions"), (["in"], c_ulong, "stfMask")),
        COMMETHOD([], HRESULT, "GetFolderType", (["out"], POINTER(GUID), "pftid")),
        COMMETHOD([], HRESULT, "SetFolderType", (["in"], POINTER(GUID), "ftid")),
        COMMETHOD([], HRESULT, "GetIcon", (["out"], POINTER(LPWSTR), "ppszIcon")),
        COMMETHOD([], HRESULT, "SetIcon", (["in"], LPCWSTR, "pszIcon")),
        COMMETHOD([], HRESULT, "Commit"),
        COMMETHOD(
            [],
            HRESULT,
            "Save",
            (["in"], POINTER(IShellItem), "psiFolderToSaveIn"),
            (["in"], LPCWSTR, "pszLibraryName"),
            (["in"], c_ulong, "lrf"),
            (["out"], POINTER(POINTER(IShellItem)), "ppsiNewItem"),
        ),
        COMMETHOD(
            [],
            HRESULT,
            "SaveInKnownFolder",
            (["in"], POINTER(GUID), "kfid"),
            (["in"], LPCWSTR, "pszLibraryName"),
            (["in"], c_ulong, "lrf"),
            (["out"], POINTER(POINTER(IShellItem)), "ppsiNewItem"),
        ),
    ]
    QueryInterface: Callable[[GUID, comtypes.IUnknown], HRESULT]
    AddRef: Callable[[], ULONG]
    Release: Callable[[], ULONG]
    LoadLibraryFromItem: Callable[[_Pointer[IShellLibrary], IShellItem, c_ulong], HRESULT]
    LoadLibraryFromKnownFolder: Callable[[_Pointer[IShellLibrary], GUID, c_ulong], HRESULT]
    AddFolder: Callable[[_Pointer[IShellLibrary], IShellItem], HRESULT]
    RemoveFolder: Callable[[_Pointer[IShellLibrary], IShellItem], HRESULT]
    GetFolders: Callable[[_Pointer[IShellLibrary], c_int, GUID, _Pointer[_Pointer[c_void_p]]], HRESULT]
    ResolveFolder: Callable[[_Pointer[IShellLibrary], IShellItem, c_ulong, GUID, _Pointer[_Pointer[c_void_p]]], HRESULT]
    GetDefaultSaveFolder: Callable[[_Pointer[IShellLibrary], c_int, GUID, _Pointer[_Pointer[c_void_p]]], HRESULT]
    SetDefaultSaveFolder: Callable[[_Pointer[IShellLibrary], c_int, IShellItem], HRESULT]
    GetOptions: Callable[[_Pointer[IShellLibrary], _Pointer[c_uint]], HRESULT]
    SetOptions: Callable[[_Pointer[IShellLibrary], c_ulong, c_ulong], HRESULT]
    GetFolderType: Callable[[_Pointer[IShellLibrary], GUID], HRESULT]
    SetFolderType: Callable[[_Pointer[IShellLibrary], GUID], HRESULT]
    GetIcon: Callable[[_Pointer[IShellLibrary], _Pointer[LPWSTR]], HRESULT]
    SetIcon: Callable[[_Pointer[IShellLibrary], LPCWSTR], HRESULT]
    Commit: Callable[[_Pointer[IShellLibrary]], HRESULT]
    Save: Callable[[_Pointer[IShellLibrary], IShellItem, LPCWSTR, c_ulong, IShellItem], HRESULT]
    SaveInKnownFolder: Callable[[_Pointer[IShellLibrary], GUID, LPCWSTR, c_ulong, IShellItem], HRESULT]


class ShellLibrary(comtypes.COMObject):
    _com_interfaces_: Sequence[type[comtypes.IUnknown]] = [IShellLibrary]

    def LoadLibraryFromItem(self, psi: IShellItem, grfMode: c_ulong | int) -> HRESULT:  # noqa: N803
        return S_OK

    def LoadLibraryFromKnownFolder(self, kfidLibrary: GUID, grfMode: c_ulong | int) -> HRESULT:  # noqa: N803
        return S_OK

    def AddFolder(self, psi: IShellItem) -> HRESULT:
        return S_OK

    def RemoveFolder(self, psi: IShellItem) -> HRESULT:
        return S_OK

    def GetFolders(self, lff: c_int | int, riid: GUID, ppv: _Pointer[_Pointer[c_void_p]]) -> HRESULT:
        return S_OK

    def ResolveFolder(self, psi: IShellItem, grfMode: c_ulong | int, riid: GUID, ppv: _Pointer[_Pointer[c_void_p]]) -> HRESULT:  # noqa: N803
        return S_OK

    def GetDefaultSaveFolder(self, dsft: c_int | int, riid: GUID, ppv: _Pointer[_Pointer[c_void_p]]) -> HRESULT:
        return S_OK

    def SetDefaultSaveFolder(self, dsft: c_int | int, psi: IShellItem) -> HRESULT:
        return S_OK

    def GetOptions(self, pOptions: _Pointer[c_uint]) -> HRESULT:  # noqa: N803
        return S_OK

    def SetOptions(self, stfOptions: c_ulong | int, stfMask: c_ulong | int) -> HRESULT:  # noqa: N803
        return S_OK

    def GetFolderType(self, pftid: GUID) -> HRESULT:
        return S_OK

    def SetFolderType(self, ftid: GUID) -> HRESULT:
        return S_OK

    def GetIcon(self, ppszIcon: _Pointer[LPWSTR]) -> HRESULT:  # noqa: N803
        return S_OK

    def SetIcon(self, pszIcon: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def Commit(self) -> HRESULT:
        return S_OK

    def Save(self, psiFolderToSaveIn: IShellItem, pszLibraryName: LPCWSTR | str, lrf: c_ulong | int, ppsiNewItem: IShellItem) -> HRESULT:  # noqa: N803
        return S_OK

    def SaveInKnownFolder(self, kfid: GUID, pszLibraryName: LPCWSTR | str, lrf: c_ulong | int, ppsiNewItem: IShellItem) -> HRESULT:  # noqa: N803
        return S_OK


class IFileOpenDialog(IFileDialog):
    _case_insensitive_: bool = True
    _iid_: GUID = IID_IFileOpenDialog
    _methods_: ClassVar[list[_ComMemberSpec]] = [
        COMMETHOD([], HRESULT, "GetResults", (["out"], POINTER(POINTER(IShellItemArray)), "ppenum")),
        COMMETHOD([], HRESULT, "GetSelectedItems", (["out"], POINTER(POINTER(IShellItemArray)), "ppsai")),
    ]
    GetResults: Callable[[], IShellItemArray]
    GetSelectedItems: Callable[[], IShellItemArray]


class FileOpenDialog(comtypes.COMObject):
    _com_interfaces_: Sequence[type[comtypes.IUnknown]] = [IFileOpenDialog]

    def QueryInterface(self, riid: GUID, ppvObject: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK

    def AddRef(self) -> ULONG:
        return ULONG(-1)

    def Release(self) -> ULONG:
        return ULONG(-1)

    def Show(self, hwndParent: HWND | int) -> HRESULT:  # noqa: N803
        return S_OK

    def SetFileTypes(self, cFileTypes: c_uint | int, rgFilterSpec: Array[COMDLG_FILTERSPEC]) -> HRESULT:  # noqa: N803
        return S_OK

    def SetFileTypeIndex(self, iFileType: c_uint | int) -> HRESULT:  # noqa: N803
        return S_OK

    def GetFileTypeIndex(self, piFileType: _Pointer[c_uint]) -> HRESULT:  # noqa: N803
        return S_OK

    def Advise(self, pfde: _Pointer[comtypes.IUnknown], pdwCookie: _Pointer[DWORD]) -> HRESULT:  # noqa: N803
        return S_OK

    def Unadvise(self, dwCookie: int) -> HRESULT:  # noqa: N803
        return S_OK

    def SetOptions(self, fos: int) -> HRESULT:
        return S_OK

    def GetOptions(self, pfos: _Pointer[DWORD]) -> HRESULT:
        return S_OK

    def SetDefaultFolder(self, psi: IShellItem) -> HRESULT:
        return S_OK

    def SetFolder(self, psi: IShellItem) -> HRESULT:
        return S_OK

    def GetFolder(self, ppsi: IShellItem) -> HRESULT:
        return S_OK

    def GetCurrentSelection(self, ppsi: IShellItem) -> HRESULT:
        return S_OK

    def SetFileName(self, pszName: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def GetFileName(self, pszName: _Pointer[LPWSTR]) -> HRESULT:  # noqa: N803
        return S_OK

    def SetTitle(self, pszTitle: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def SetOkButtonLabel(self, pszText: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def SetFileNameLabel(self, pszLabel: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def GetResult(self, ppsi: IShellItem) -> HRESULT:
        return S_OK

    def AddPlace(self, psi: IShellItem, fdap: c_int) -> HRESULT:
        return S_OK

    def SetDefaultExtension(self, pszDefaultExtension: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def Close(self, hr: HRESULT | int) -> HRESULT:
        return S_OK

    def SetClientGuid(self, guid: GUID) -> HRESULT:
        return S_OK

    def ClearClientData(self) -> HRESULT:
        return S_OK

    def SetFilter(self, isFilter: IShellItemFilter) -> HRESULT:  # noqa: N803
        return S_OK

    def GetResults(self, isArray: IShellItemArray) -> HRESULT:  # noqa: N803
        return S_OK

    def GetSelectedItems(self, ppsai: IShellItemArray) -> HRESULT:
        return S_OK


class IFileSaveDialog(IFileDialog):
    _case_insensitive_: bool = True
    _iid_: GUID = IID_IFileSaveDialog
    _methods_: ClassVar[list[_ComMemberSpec]] = [
        COMMETHOD([], HRESULT, "SetSaveAsItem", (["in"], POINTER(IShellItem), "psi")),
        COMMETHOD([], HRESULT, "SetProperties", (["in"], POINTER(comtypes.IUnknown), "pStore")),
        COMMETHOD([], HRESULT, "SetCollectedProperties", (["in"], POINTER(comtypes.IUnknown), "pList"), (["in"], BOOL, "fAppendDefault")),
        COMMETHOD([], HRESULT, "GetProperties", (["out"], POINTER(POINTER(comtypes.IUnknown)), "ppStore")),
        COMMETHOD(
            [],
            HRESULT,
            "ApplyProperties",
            (["in"], POINTER(IShellItem), "psi"),
            (["in"], POINTER(comtypes.IUnknown), "pStore"),
            (["in"], HWND, "hwnd"),
            (["in"], POINTER(comtypes.IUnknown), "pSink"),
        ),
    ]
    SetSaveAsItem: Callable[[IShellItem], HRESULT]
    SetProperties: Callable[[_Pointer[comtypes.IUnknown]], HRESULT]
    SetCollectedProperties: Callable[[_Pointer[comtypes.IUnknown], BOOL], HRESULT]
    GetProperties: Callable[[_Pointer[_Pointer[comtypes.IUnknown]]], HRESULT]
    ApplyProperties: Callable[[IShellItem, _Pointer[comtypes.IUnknown], HWND, _Pointer[comtypes.IUnknown]], HRESULT]


class FileSaveDialog(comtypes.COMObject):
    _com_interfaces_: Sequence[type[comtypes.IUnknown]] = [IFileSaveDialog]

    def QueryInterface(self, riid: GUID, ppvObject: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK

    def AddRef(self) -> ULONG:
        return ULONG(-1)

    def Release(self) -> ULONG:
        return ULONG(-1)

    def Show(self, hwndParent: HWND | int) -> HRESULT:  # noqa: N803
        return S_OK

    def SetFileTypes(self, cFileTypes: c_uint | int, rgFilterSpec: _Pointer[COMDLG_FILTERSPEC]) -> HRESULT:  # noqa: N803
        return S_OK

    def SetFileTypeIndex(self, iFileType: c_uint | int) -> HRESULT:  # noqa: N803
        return S_OK

    def GetFileTypeIndex(self, piFileType: _Pointer[c_uint]) -> HRESULT:  # noqa: N803
        return S_OK

    def Advise(self, pfde: _Pointer[comtypes.IUnknown], pdwCookie: _Pointer[DWORD]) -> HRESULT:  # noqa: N803
        return S_OK

    def Unadvise(self, dwCookie: int) -> HRESULT:  # noqa: N803
        return S_OK

    def SetOptions(self, fos: int) -> HRESULT:
        return S_OK

    def GetOptions(self, pfos: _Pointer[DWORD]) -> HRESULT:
        return S_OK

    def SetDefaultFolder(self, psi: IShellItem) -> HRESULT:
        return S_OK

    def SetFolder(self, psi: IShellItem) -> HRESULT:
        return S_OK

    def GetFolder(self, ppsi: IShellItem) -> HRESULT:
        return S_OK

    def GetCurrentSelection(self, ppsi: IShellItem) -> HRESULT:
        return S_OK

    def SetFileName(self, pszName: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def GetFileName(self, pszName: _Pointer[LPWSTR]) -> HRESULT:  # noqa: N803
        return S_OK

    def SetTitle(self, pszTitle: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def SetOkButtonLabel(self, pszText: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def SetFileNameLabel(self, pszLabel: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def GetResult(self, ppsi: IShellItem) -> HRESULT:
        return S_OK

    def AddPlace(self, psi: IShellItem, fdap: c_int) -> HRESULT:
        return S_OK

    def SetDefaultExtension(self, pszDefaultExtension: LPCWSTR | str) -> HRESULT:  # noqa: N803
        return S_OK

    def Close(self, hr: HRESULT | int) -> HRESULT:
        return S_OK

    def SetClientGuid(self, guid: GUID) -> HRESULT:
        return S_OK

    def ClearClientData(self) -> HRESULT:
        return S_OK

    def SetFilter(self, pFilter: IShellItemFilter) -> HRESULT:  # noqa: N803
        return S_OK

    def SetSaveAsItem(self, psi: IShellItem) -> HRESULT:
        return S_OK

    def SetProperties(self, pStore: _Pointer[comtypes.IUnknown]) -> HRESULT:  # noqa: N803
        return S_OK

    def SetCollectedProperties(self, pList: _Pointer[comtypes.IUnknown], fAppendDefault: BOOL | int) -> HRESULT:  # noqa: N803
        return S_OK

    def GetProperties(self, ppStore: comtypes.IUnknown) -> HRESULT:  # noqa: N803
        return S_OK

    def ApplyProperties(self, psi: IShellItem, pStore: _Pointer[comtypes.IUnknown], hwnd: HWND | int, pSink: _Pointer[comtypes.IUnknown]) -> HRESULT:  # noqa: N803
        return S_OK


class IFileDialogCustomize(comtypes.IUnknown):
    _case_insensitive_ = True
    _iid_ = IID_IFileDialogCustomize
    _methods_: ClassVar[list[_ComMemberSpec]] = [
        COMMETHOD([], HRESULT, "EnableOpenDropDown", (["in"], c_uint, "dwIDCtl")),
        COMMETHOD([], HRESULT, "AddText", (["in"], c_uint, "dwIDCtl"), (["in"], LPCWSTR, "pszText")),
        COMMETHOD([], HRESULT, "AddPushButton", (["in"], c_uint, "dwIDCtl"), (["in"], LPCWSTR, "pszLabel")),
        COMMETHOD([], HRESULT, "AddCheckButton", (["in"], c_uint, "dwIDCtl"), (["in"], LPCWSTR, "pszLabel"), (["in"], c_int, "bChecked")),
        COMMETHOD([], HRESULT, "AddRadioButtonList", (["in"], c_uint, "dwIDCtl")),
        COMMETHOD([], HRESULT, "AddComboBox", (["in"], c_uint, "dwIDCtl")),
        COMMETHOD([], HRESULT, "AddControlItem", (["in"], c_uint, "dwIDCtl"), (["in"], c_uint, "dwIDItem"), (["in"], LPCWSTR, "pszLabel")),
        COMMETHOD([], HRESULT, "AddEditBox", (["in"], c_uint, "dwIDCtl"), (["in"], LPCWSTR, "pszText")),
        COMMETHOD([], HRESULT, "AddSeparator", (["in"], c_uint, "dwIDCtl")),
        COMMETHOD([], HRESULT, "AddMenu", (["in"], c_uint, "dwIDCtl"), (["in"], LPCWSTR, "pszLabel")),
        COMMETHOD([], HRESULT, "SetControlLabel", (["in"], c_uint, "dwIDCtl"), (["in"], LPCWSTR, "pszLabel")),
        COMMETHOD([], HRESULT, "SetControlState", (["in"], c_uint, "dwIDCtl"), (["in"], c_int, "dwState")),
        COMMETHOD([], HRESULT, "SetCheckButtonState", (["in"], c_uint, "dwIDCtl"), (["in"], c_int, "bChecked")),
        COMMETHOD([], HRESULT, "GetCheckButtonState", (["in"], c_uint, "dwIDCtl"), (["out"], POINTER(c_int), "pbChecked")),
        COMMETHOD([], HRESULT, "SetEditBoxText", (["in"], c_uint, "dwIDCtl"), (["in"], LPCWSTR, "pszText")),
        COMMETHOD([], HRESULT, "GetEditBoxText", (["in"], c_uint, "dwIDCtl"), (["out"], POINTER(LPCWSTR), "ppszText")),
        COMMETHOD([], HRESULT, "SetControlItemText", (["in"], c_uint, "dwIDCtl"), (["in"], c_uint, "dwIDItem"), (["in"], LPCWSTR, "pszLabel")),
        COMMETHOD([], HRESULT, "GetControlItemState", (["in"], c_uint, "dwIDCtl"), (["in"], c_uint, "dwIDItem"), (["out"], POINTER(c_int), "pdwState")),
        COMMETHOD([], HRESULT, "SetControlItemState", (["in"], c_uint, "dwIDCtl"), (["in"], c_uint, "dwIDItem"), (["in"], c_int, "dwState")),
        COMMETHOD([], HRESULT, "GetSelectedControlItem", (["in"], c_uint, "dwIDCtl"), (["out"], POINTER(c_uint), "pdwIDItem")),
        COMMETHOD([], HRESULT, "SetSelectedControlItem", (["in"], c_uint, "dwIDCtl"), (["in"], c_uint, "dwIDItem")),
        COMMETHOD([], HRESULT, "StartVisualGroup", (["in"], c_uint, "dwIDCtl"), (["in"], LPCWSTR, "pszLabel")),
        COMMETHOD([], HRESULT, "EndVisualGroup", (["in"], c_uint, "dwIDCtl")),
        COMMETHOD([], HRESULT, "MakeProminent", (["in"], c_uint, "dwIDCtl")),
        COMMETHOD([], HRESULT, "RemoveControlItem", (["in"], c_uint, "dwIDCtl"), (["in"], c_uint, "dwIDItem")),
        COMMETHOD([], HRESULT, "RemoveAllControlItems", (["in"], c_uint, "dwIDCtl")),
        COMMETHOD([], HRESULT, "GetControlState", (["in"], c_uint, "dwIDCtl"), (["out"], POINTER(c_int), "pdwState")),
    ]
    EnableOpenDropDown: Callable[[int], HRESULT]
    AddText: Callable[[int, str], HRESULT]
    AddPushButton: Callable[[int, str], HRESULT]
    AddCheckButton: Callable[[int, str, int], HRESULT]
    AddRadioButtonList: Callable[[int], HRESULT]
    AddComboBox: Callable[[int], HRESULT]
    AddControlItem: Callable[[int, int, str], HRESULT]
    AddEditBox: Callable[[int, str], HRESULT]
    AddSeparator: Callable[[int], HRESULT]
    AddMenu: Callable[[int, str], HRESULT]
    SetControlLabel: Callable[[int, str], HRESULT]
    SetControlState: Callable[[int, int], HRESULT]
    SetCheckButtonState: Callable[[int, int], HRESULT]
    GetCheckButtonState: Callable[[int], int]
    SetEditBoxText: Callable[[int, str], HRESULT]
    GetEditBoxText: Callable[[int], LPCWSTR]
    SetControlItemText: Callable[[int, int, str], HRESULT]
    GetControlItemState: Callable[[int, int], int]
    SetControlItemState: Callable[[int, int, int], HRESULT]
    GetSelectedControlItem: Callable[[int], int]
    SetSelectedControlItem: Callable[[int, int], HRESULT]
    StartVisualGroup: Callable[[int, str], HRESULT]
    EndVisualGroup: Callable[[int], HRESULT]
    MakeProminent: Callable[[int], HRESULT]
    RemoveControlItem: Callable[[int, int], HRESULT]
    RemoveAllControlItems: Callable[[int], HRESULT]
    GetControlState: Callable[[int], int]


class IFileDialogControlEvents(comtypes.IUnknown):
    _case_insensitive_ = True
    _iid_ = IID_IFileDialogControlEvents
    _methods_: ClassVar[list[_ComMemberSpec]] = [
        COMMETHOD([], HRESULT, "OnItemSelected", (["in"], comtypes.POINTER(IFileDialogCustomize), "pfdc"), (["in"], c_int, "dwIDCtl"), (["in"], c_int, "dwIDItem")),
        COMMETHOD([], HRESULT, "OnButtonClicked", (["in"], comtypes.POINTER(IFileDialogCustomize), "pfdc"), (["in"], c_int, "dwIDCtl")),
        COMMETHOD([], HRESULT, "OnCheckButtonToggled", (["in"], comtypes.POINTER(IFileDialogCustomize), "pfdc"), (["in"], c_int, "dwIDCtl"), (["in"], c_bool, "bChecked")),
        COMMETHOD([], HRESULT, "OnControlActivating", (["in"], comtypes.POINTER(IFileDialogCustomize), "pfdc"), (["in"], c_int, "dwIDCtl")),
    ]
    OnButtonClicked: Callable[[IFileDialogCustomize, c_uint], HRESULT]
    OnCheckButtonToggled: Callable[[IFileDialogCustomize, c_uint, c_int], HRESULT]
    OnControlActivating: Callable[[IFileDialogCustomize, c_uint], HRESULT]
    OnItemSelected: Callable[[IFileDialogCustomize, c_uint, c_uint], HRESULT]


if __name__ == "__main__":
    assert ModalWindow()
    assert ShellItem()
    assert ContextMenu()
    assert ShellFolder()
    assert ShellItemArray()
    assert ShellItemFilter()
    assert EnumShellItems()
    assert PropertyStore()
    assert FileOperationProgressSink()
    assert FileDialogEvents()
    assert FileDialog()
    assert ShellLibrary()
    assert FileOpenDialog()
    assert FileSaveDialog()



from __future__ import annotations

import hashlib
import inspect
import json
import re
import secrets
import shutil
import tempfile
import time

from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable

from utility.misc import ensure_directory_exists

# Handle optional certifi dependency
try:
    import certifi  # pyright: ignore[reportMissingImports]
except ImportError:
    if not TYPE_CHECKING:
        certifi = None  # type: ignore[assignment, unused-ignore]

try:
    import requests
except ImportError:
    if not TYPE_CHECKING:
        requests = None  # type: ignore[assignment, unused-ignore]

try:
    import urllib3
except ImportError:
    if not TYPE_CHECKING:
        urllib3 = None  # type: ignore[assignment, unused-ignore]

from loggerplus import RobustLogger

try:
    from Crypto.Cipher import AES
    from Crypto.Util import Counter

    _CRYPTO_AVAILABLE = True
except ImportError:
    _CRYPTO_AVAILABLE = False
    AES = None  # type: ignore[assignment, misc]
    Counter = None  # type: ignore[assignment, misc]

from utility.updater.crypto import (
    a32_to_str,
    base64_to_a32,
    base64_url_decode,
    decrypt_mega_attr,
    get_chunks,
    str_to_a32,
)

if TYPE_CHECKING:
    import os

    from logging import Logger

    from typing_extensions import Literal


class FileDownloaderError(ConnectionError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class FileDownloader:
    """The FileDownloader object downloads files to memory and
    verifies their hash.  If hash is verified data is either
    written to disk to returned to calling object.

    Args:
    ----
        filepath (os.PathLike | str): The filepath to download to.
        urls (list[str]): List of urls to use for file download
        hexdigest (str | None): The hash of the file to download

    Kwargs:
    ------
        progress_hooks ( list[Callable[dict[str, Any]]] ): A list of callable functions that can be used when we report the progress. Untested.
        headers: (dict[str, Any] | None): custom headers to pass
        max_download_retries (int): Maximum number of times to attempt to download the file. Defaults to zero.
        verify (bool):
            True: Verify https connection
            False: Do not verify https connection
        http_timeout: Unknown but defaults to None.
    """

    def __init__(
        self,
        filepath: os.PathLike | str,
        urls: list[str],
        hexdigest: str | None,
        *,
        progress_hooks: list[Callable[[dict[str, Any]], Any]] | None = None,
        headers: dict[str, Any] | None = None,
        max_download_retries: int | None = None,
        verify: bool = True,
        http_timeout: int | None = None,
        logger: Logger | None = None,
    ):
        # We'll append the filename to one of the provided urls
        # to create the download link
        if not filepath:
            raise FileDownloaderError("No filename provided", expected=True)
        self.filepath = Path(filepath)
        self.log = logger or RobustLogger()

        self.file_binary_data: list = []  # Hold all binary data once file has been downloaded
        self.file_binary_path: Path = Path(f"{self.filepath}.part")  # Temporary file to hold large download data

        if not urls:
            raise FileDownloaderError("No urls provided", expected=True)
        if not isinstance(urls, list):  # User may have accidentally passed a string to the urls parameter
            raise FileDownloaderError("Must pass list of urls", expected=True)
        self.urls: list[str] = urls

        self.hexdigest = hexdigest
        self.verify = verify  # Specify if we want to verify TLS connections
        self.max_download_retries: int = max_download_retries or 0  # Max attempts to download resource
        self.progress_hooks: list[Callable[[dict[str, Any]], Any]] = progress_hooks or []  # Progress hooks to be called
        self.block_size: int = 4096 * 4  # Initial block size for each read
        self.file_binary_type: Literal["memory", "file"] = "memory"  # Storage type
        self.downloaded_filename: str = self.filepath.name

        # Extra headers
        self.headers: dict[str, Any] = headers or {"User-Agent": "MyAppName/1.0 (https://myappwebsite.com/)"}
        self.http_timeout = http_timeout
        self.download_max_size: int = 16 * 1024 * 1024  # Max size of download to memory, larger file will be stored to file
        self.content_length: int | None = None  # Total length of data to download.
        self.session = requests.Session()
        self.session.headers.update(self.headers)

        self.http_pool = self._get_http_pool(secure=self.verify)

    def _get_http_pool(self, *, secure=True):
        if secure:
            if certifi is None:
                # certifi not available, use system default certificates or disable verification
                self.log.warning("certifi not available, using system default certificate store")
                _http = urllib3.PoolManager(
                    cert_reqs="CERT_REQUIRED",
                    timeout=self.http_timeout,
                )
            else:
                _http = urllib3.PoolManager(
                    cert_reqs="CERT_REQUIRED",
                    ca_certs=certifi.where(),
                    timeout=self.http_timeout,
                )
        else:
            _http = urllib3.PoolManager(timeout=self.http_timeout)

        if self.headers:
            self._apply_custom_headers(_http)
        self.log.debug(f"HTTP Timeout is {self.http_timeout!s}")  # noqa: G004
        return _http

    def _apply_custom_headers(self, _http: urllib3.PoolManager):
        urllib_keys = inspect.getfullargspec(urllib3.make_headers).args
        urllib_headers = {header: value for header, value in self.headers.items() if header in urllib_keys}
        other_headers = {header: value for header, value in self.headers.items() if header not in urllib_keys}
        _headers: dict[str, str] = urllib3.make_headers(**urllib_headers)
        _headers.update(other_headers)
        if not isinstance(_http.headers, dict):
            _http.headers = dict(_http.headers)
        _http.headers.update(_headers)

    def _start_hooks(self, content_length):
        for hook in self.progress_hooks:
            hook({"action": "starting", "data": {"total": content_length}})

    def _progress_hooks(self, just_downloaded, total):
        for hook in self.progress_hooks:
            hook({"action": "update_progress", "data": {"downloaded": just_downloaded, "total": total}})

    @staticmethod
    def _get_filename_from_cd(cd):
        """Get filename from content-disposition header."""
        if not cd:
            return None
        fname = re.findall('filename="(.+)"', cd)
        return fname[0] if fname else None

    def download_verify_write(self) -> bool:
        """Downloads file then verifies against provided hash
        If hash verfies then writes data to disk.

        Returns:
             (bool):
                 True - Hashes match or no hash was given during initialization.
                 False - Hashes don't match
        """
        success: bool = False
        for url in self.urls:
            try:
                with self.session.get(url, stream=True, timeout=self.http_timeout, verify=self.verify) as r:
                    r.raise_for_status()

                    # Determine the filename from the Content-Disposition header or URL.
                    filename = self._get_filename_from_cd(r.headers.get("Content-Disposition")) or Path(url).name
                    self.downloaded_filename = filename
                    RobustLogger().info(f"Expected downloaded filename: {self.downloaded_filename}")
                    file_path = self.filepath.parent / filename
                    RobustLogger().info(f"Expected download path: {file_path}")

                    # Start the download process.
                    content_length = int(r.headers.get("Content-Length", 0))
                    self._start_hooks(content_length)
                    start_time = time.time()
                    with file_path.open("wb") as f:
                        chunk_start = 0
                        for chunk in r.iter_content(chunk_size=8192):
                            if not chunk:
                                continue
                            f.write(chunk)
                            chunk_start += len(chunk)
                            for hook in self.progress_hooks:
                                hook(
                                    {
                                        "action": "update_progress",
                                        "data": {
                                            "total": content_length,
                                            "downloaded": chunk_start,
                                            "status": "downloading",
                                            "percent_complete": self._calc_progress_percent(chunk_start, content_length),
                                            "time": self._calc_eta(start_time, time.time(), content_length, chunk_start),
                                        },
                                    }
                                )
                    success = self._check_hash()
                    if success:
                        break
            except requests.exceptions.RequestException:
                self.log.exception("Request failed")
                continue
        return success

    @staticmethod
    def _best_block_size(elapsed_time: float, _bytes: float) -> int:
        new_max = min(max(_bytes * 2.0, 1.0), 4194304)  # Do not surpass 4 MB
        one_millisecond = 0.001
        if elapsed_time < one_millisecond:
            return int(new_max)
        rate = _bytes / elapsed_time
        if rate > new_max:
            return int(new_max)
        # Returns best block size for current Internet connection speed
        new_min = max(_bytes / 2.0, 1.0)
        return int(new_min) if rate < new_min else int(rate)

    def _check_hash(self) -> bool:
        if not self.hexdigest:
            return True
        sha256 = hashlib.sha256()
        with self.filepath.open("rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                sha256.update(chunk)
        return sha256.hexdigest() == self.hexdigest

    def _write_to_file(self):
        # Writes download data to disk
        if self.file_binary_type == "memory":
            with self.filepath.open("wb") as f:
                for block in self.file_binary_data:
                    f.write(block)
        else:
            filepath = Path(self.filepath)
            if filepath.exists():
                filepath.unlink(missing_ok=True)
            self.file_binary_path.rename(self.filepath)

    @staticmethod
    def _get_content_length(
        data,
    ) -> int | None:
        content_length_lookup: str | None = data.headers.get("Content-Length")
        log = RobustLogger()
        log.debug("Got content length of: %s", content_length_lookup)
        return int(content_length_lookup) if content_length_lookup else None

    @staticmethod
    def _calc_eta(
        start: float,
        now: float,
        total: int | None,
        current: int,
    ) -> str:
        """Calculates remaining time of download."""
        if total is None:
            return "--:--"

        diff = now - start
        one_millisecond = 0.001
        if current == 0 or diff < one_millisecond:
            return "--:--"

        rate = float(current) / diff
        eta = int((float(total) - float(current)) / rate)
        (eta_mins, eta_secs) = divmod(eta, 60)
        long_time = 99
        return "--:--" if eta_mins > long_time else f"{eta_mins:02d}:{eta_secs:02d}"

    @staticmethod
    def _calc_progress_percent(
        received: float | str,
        total: int | None,
    ) -> str:
        if total is None:
            return "-.-%"
        percent = float(received) / total * 100
        return f"{percent:.1f}"


def _api_request(data, sequence_num: int | None = None, **kwargs):
    if sequence_num is None:
        sequence_num = secrets.randbelow(0xFFFFFFFF)
    params: dict[str, Any] = {"id": sequence_num}
    sequence_num += 1

    # if self.sid:
    #     params["sid"] = self.sid

    if kwargs:
        params.update(kwargs)

    # Ensure input data is a list
    if not isinstance(data, list):
        data = [data]

    req = requests.post("https://g.api.mega.co.nz/cs", params=params, data=json.dumps(data), timeout=160)
    json_resp = json.loads(req.text)

    # if numeric error code response
    if isinstance(json_resp, int):
        raise ConnectionRefusedError(f"Api request error: {json_resp}")
    return json_resp[0]


def _download_file(
    file_handle: str,
    file_key: str,
    dest: os.PathLike | str | None = None,
    dest_filename: str | None = None,
    is_public: bool = False,
    file: dict[str, Any] | None = None,
    progress_hooks: list[Callable[[dict[str, Any]], Any]] | None = None,
):
    if not _CRYPTO_AVAILABLE:
        raise ImportError("pycryptodome is required for MEGA file downloads. Install it with: pip install pycryptodome")
    dest_path = Path(dest or Path.cwd()).absolute()
    if file is None:
        if is_public:
            file_key = base64_to_a32(file_key)
            file_data = _api_request({"a": "g", "g": 1, "p": file_handle})
        else:
            file_data = _api_request({"a": "g", "g": 1, "n": file_handle})

        k = (file_key[0] ^ file_key[4], file_key[1] ^ file_key[5], file_key[2] ^ file_key[6], file_key[3] ^ file_key[7])
        iv = file_key[4:6] + (0, 0)
        meta_mac = file_key[6:8]
    else:
        file_data = _api_request({"a": "g", "g": 1, "n": file["h"]})
        k = file["k"]
        iv = file["iv"]
        meta_mac = file["meta_mac"]

    # Seems to happens sometime... When  this occurs, files are
    # inaccessible also in the official also in the official web app.
    # Strangely, files can come back later.
    if "g" not in file_data:
        raise ConnectionError("File not accessible anymore")
    file_url = file_data["g"]
    file_size = file_data["s"]
    attribs = base64_url_decode(file_data["at"])
    decrypted_attribs = decrypt_mega_attr(attribs, k)
    if decrypted_attribs is False:
        raise ValueError("Could not retrieve attribs?")

    file_name = decrypted_attribs["n"] if dest_filename is None else dest_filename
    input_file = requests.get(file_url, stream=True, timeout=30).raw

    temp_output_file = tempfile.NamedTemporaryFile(mode="w+b", prefix="megapy_", delete=False)

    k_str = a32_to_str(k)
    counter = Counter.new(
        128,
        initial_value=((iv[0] << 32) + iv[1]) << 64,
    )
    aes = AES.new(k_str, AES.MODE_CTR, counter=counter)

    mac_str = b"\0" * 16
    mac_encryptor = AES.new(k_str, AES.MODE_CBC, mac_str)
    iv_str = a32_to_str([iv[0], iv[1], iv[0], iv[1]])

    start_download = time.time()

    for chunk_start, chunk_size in get_chunks(file_size):
        chunk = input_file.read(chunk_size)
        chunk = aes.decrypt(chunk)
        temp_output_file.write(chunk)

        encryptor = AES.new(k_str, AES.MODE_CBC, iv_str)
        for i in range(0, len(chunk) - 16, 16):
            block = chunk[i : i + 16]
            encryptor.encrypt(block)

        # Fix for files under 16 bytes failing
        if file_size > 16:
            i += 16
        else:
            i = 0

        block = chunk[i : i + 16]
        if len(block) % 16:
            block += b"\0" * (16 - (len(block) % 16))
        mac_str = mac_encryptor.encrypt(encryptor.encrypt(block))

        # Temp file size
        file_info = Path(temp_output_file.name).stat()

        # If content length is None we will return a static percent
        # -.-%
        percent = FileDownloader._calc_progress_percent(chunk_start, file_size)

        # If content length is None we will return a static time remaining
        # --:--
        time_left = FileDownloader._calc_eta(start_download, time.time(), file_size, chunk_start)

        status = {
            "action": "update_progress",
            "data": {
                "total": file_size,
                "downloaded": chunk_start,
                "status": "downloading",
                "percent_complete": percent,
                "time": time_left,
            },
        }

        log = RobustLogger()

        # Call all progress hooks with status data
        log.debug(status)
        if progress_hooks is not None:
            for ph in progress_hooks:
                try:
                    ph(status)
                except Exception as err:  # noqa: PERF203
                    log.exception("Exception in callback '%s': %s", ph.__name__, err)
        log.debug(f"Status - {file_info.st_size / file_size * 100:.2f} downloaded")  # noqa: G004
        log.debug(f"{file_info.st_size} of {file_size} downloaded")  # noqa: G004

    file_mac = str_to_a32(mac_str)

    temp_output_file.close()

    # check mac integrity
    if (file_mac[0] ^ file_mac[1], file_mac[2] ^ file_mac[3]) != meta_mac:
        raise ValueError("Mismatched mac")

    if not file_name:
        file_name = dest_path.name
    if dest_path.name == file_name:
        dest_path = dest_path.parent
    dest_filepath = dest_path / file_name
    if not dest_filepath.parent.is_dir():
        ensure_directory_exists(dest_filepath.parent)
    shutil.move(temp_output_file.name, dest_filepath)


def download_mega_file_url(
    url: str,
    dest_path: os.PathLike | str | None = None,
    dest_filename: str | None = None,
    progress_hooks: list[Callable] | None = None,
):
    """Download a file by it's public mega url."""
    # Splitting the URL at the first occurrence of '/file/' and '#'
    base_url, rest = url.split("/file/")
    file_id, decryption_key = rest.split("#")

    # Reconstructing the base URL to include '/file/'
    base_url = f"{base_url}/file/"

    print("Base URL:", base_url)
    print("File ID:", file_id)
    print("Decryption Key:", decryption_key)
    _download_file(file_id, decryption_key, dest_path, dest_filename, is_public=True, progress_hooks=progress_hooks)



"""TSLPatcher diff module - tools for analyzing and generating patch data.

This module provides:
- analyzers: Convert raw diffs into TSLPatcher modification structures
- cache: Diff result caching system
- engine: Core comparison operations for diff detection
- enhanced_engine: Enhanced diff engine with formatting support
- formatters: Output formatters for different diff formats
- generator: Generate complete tslpatchdata folder structures
- objects: Structured diff result objects
- resolution: Resource resolution order handling for installation comparisons
- structured: Detailed diff results for INI generation
"""

from __future__ import annotations

from pykotor.tslpatcher.diff.analyzers import DiffAnalyzerFactory
from pykotor.tslpatcher.diff.cache import DiffCache, load_diff_cache, restore_strref_cache_from_cache, save_diff_cache
from pykotor.tslpatcher.diff.engine import (
    CachedFileComparison,
    DiffContext,
    diff_capsule_files,
    diff_data,
    diff_directories,
)
from pykotor.tslpatcher.diff.enhanced_engine import EnhancedDiffEngine
from pykotor.tslpatcher.diff.formatters import (
    ContextFormatter,
    DefaultFormatter,
    DiffFormatter,
    FormatterFactory,
    SideBySideFormatter,
    UnifiedFormatter,
)
from pykotor.tslpatcher.diff.generator import (
    TSLPatchDataGenerator,
    determine_install_folders,
    validate_tslpatchdata_arguments,
)
from pykotor.tslpatcher.writer import IncrementalTSLPatchDataWriter
from pykotor.tslpatcher.diff.objects import DiffEngine, DiffFormat, DiffType
from pykotor.tslpatcher.diff.resolution import (
    ResolvedResource,
    build_resource_index,
    collect_all_resource_identifiers,
    determine_tslpatcher_destination,
    diff_installations_with_resolution,
    explain_resolution_order,
    get_location_display_name,
    resolve_resource_in_installation,
)
from pykotor.tools.reference_cache import StrRefReferenceCache


def __getattr__(name: str):
    if name in {"DiffConfig", "run_application"}:
        from pykotor.diff_tool.app import DiffConfig, run_application

        exports = {
            "DiffConfig": DiffConfig,
            "run_application": run_application,
        }
        return exports[name]
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")

__all__ = [
    "CachedFileComparison",
    "ContextFormatter",
    "DiffConfig",
    "DefaultFormatter",
    "DiffAnalyzerFactory",
    "DiffCache",
    "DiffContext",
    "DiffEngine",
    "DiffFormat",
    "DiffFormatter",
    "DiffType",
    "EnhancedDiffEngine",
    "FormatterFactory",
    "IncrementalTSLPatchDataWriter",
    "ResolvedResource",
    "SideBySideFormatter",
    "StrRefReferenceCache",
    "TSLPatchDataGenerator",
    "UnifiedFormatter",
    "build_resource_index",
    "collect_all_resource_identifiers",
    "determine_install_folders",
    "determine_tslpatcher_destination",
    "diff_capsule_files",
    "diff_data",
    "diff_directories",
    "diff_installations_with_resolution",
    "explain_resolution_order",
    "get_location_display_name",
    "load_diff_cache",
    "resolve_resource_in_installation",
    "restore_strref_cache_from_cache",
    "run_application",
    "save_diff_cache",
    "validate_tslpatchdata_arguments",
]



"""RIM package: RIM archive data and binary I/O (read_rim, write_rim, bytes_rim)."""

from __future__ import annotations
from pykotor.resource.formats.rim.rim_data import RIM, RIMResource
from pykotor.resource.formats.rim.io_rim import (
    RIMBinaryReader,
    RIMBinaryWriter,
)
from pykotor.resource.formats.rim.rim_auto import bytes_rim, read_rim, write_rim

__all__ = [
    "RIM",
    "RIMBinaryReader",
    "RIMBinaryWriter",
    "RIMResource",
    "bytes_rim",
    "read_rim",
    "write_rim",
]



"""Dialog stunt class for handling stunt models."""

from __future__ import annotations

import uuid

from typing import Any

from pykotor.common.misc import ResRef


class DLGStunt:
    """Represents a stunt model in a dialog.

    Attributes:
    ----------
    participant: "Participant" field.
    stunt_model: "StuntModel" field.
    """

    def __init__(self):
        self._hash_cache: int = uuid.uuid4().int
        self.participant: str = ""
        self.stunt_model: ResRef = ResRef.from_blank()

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return NotImplemented  # type: ignore[no-any-return]
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        return self._hash_cache

    def to_dict(self) -> dict[str, Any]:
        return {"participant": self.participant, "stunt_model": str(self.stunt_model), "_hash_cache": self._hash_cache}

    @classmethod
    def from_dict(
        cls,
        data: dict[str, Any],
        node_map: dict[str | int, Any] | None = None,  # noqa: ARG003
    ) -> DLGStunt:
        stunt: DLGStunt = cls()
        stunt.participant = data.get("participant", "")
        stunt.stunt_model = ResRef(data.get("stunt_model", ""))
        stunt._hash_cache = data.get("_hash_cache", stunt._hash_cache)  # noqa: SLF001
        return stunt



"""K1 save/load flow: 1:1 with k1_win_gog_swkotor.exe behavior.

Exhaustive Python equivalent of SaveGame @ 004b58a0, LoadGame @ 004ba640,
DoSaveGameScreenShot @ 00401080, HasEnoughDiskSpaceForSaveGame @ 004b2520,
and callees GetFreeDiskSpace @ 004067b0, CreateDirectory2 @ 004068c0,
GetDirectorySize @ 005e6650, CleanDirectory @ 00409460. Comments and labels
use exact binary names and addresses; no renaming.

K2/TSL binary (k2_win_gog_aspyr_swkotor2.exe) equivalent functions:
  Save: StallEventSaveGame @ 007b7980 (same high-level sequence as K1).
  Load: LoadGame flow FUN_007b2f00; SetLoadBarProgress 0040da90; LoadTableInfo 00701d10 @ [this+0x1f0b4];
  Load 006535d0 @ [this+0x100fc]; AddResourceDirectory 0061aff0; RemoveResourceDirectory 0061b070;
  LoadModule 007b4fd0; flags [this+0x1f254], [this+0x1f33c], [this+0x1f344].
  DoSaveGameScreenShot TSL @ 0040e0f0 (load-bar step store; screenshot in save path).
See docs/reva_roadmap/KOTOR_SAVE_LOAD_TSL_RE_REPORT.md.

Use run_save_flow(entry, game_install_path) / run_load_flow(entry, game_install_path) to
dispatch by game via pykotor.tools.heuristics.determine_game(); K1 uses this module,
K2/TSL uses save_load_flow_tsl (run_tsl_save_flow / run_tsl_load_flow).

Call graph (see docs/reva_roadmap/K1_SAVE_LOAD_CALL_GRAPH.md):
  SaveGame @ 004b58a0 callees: GetFreeDiskSpace 004067b0, CreateDirectory2 004068c0,
  CleanDirectory 00409460, DoSaveGameScreenShot 00401080, SendServerToPlayerSaveLoad_Status 0056cb80,
  SendServerToPlayerLoadBar_StartStallEvent 0056c8b0, GetDirectorySize 005e6650, GetAliasPath 005e6890,
  operator!= 005e5390 (path 0x73d71c), GetModule 005ed530, SetCameraForScreenShot 00638e80,
  AurSaveGameSnapshot 00420f20 (via DoSaveGameScreenShot), CStr 005e5670, ~CExoString 005e5c20, etc.
  LoadGame @ 004ba640 callees: SetLoadBarProgress 005edd40, SetLoadStep 005edfd0,
  AddResourceDirectory 00408800, LoadTableInfo 00565d20 @ [this+0x1b770], Load 0052ade0 @ [this+0x100fc],
  RemoveResourceDirectory 004088d0, LoadModule 004b95b0.
  HasEnoughDiskSpaceForSaveGame @ 004b2520: (free >> 14) >= 0x640 => K1_MIN_FREE_BYTES_ENGINE.
"""

from __future__ import annotations

import shutil

from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

from pykotor.common.misc import Game
from pykotor.extract.save_load_flow_tsl import run_tsl_load_flow, run_tsl_save_flow
from pykotor.tools.heuristics import determine_game

if TYPE_CHECKING:
    from pykotor.extract.savedata import SaveFolderEntry

# Game enum values that use K1 binary flow (k1_win_gog_swkotor.exe) vs TSL (k2_win_gog_aspyr_swkotor2.exe)
_GAMES_K1 = {Game.K1, Game.K1_XBOX, Game.K1_IOS, Game.K1_ANDROID}
_GAMES_K2_TSL = {Game.K2, Game.K2_XBOX, Game.K2_IOS, Game.K2_ANDROID}

# HasEnoughDiskSpaceForSaveGame @ 004b2520: (free >> 14) >= 0x640
K1_MIN_FREE_BYTES_ENGINE: int = 0x640 * (1 << 14)  # 26_214_400 (~25 MiB)

# Path/alias constants from binary (exact addresses)
K1_ALIAS_SAVE_PATH: int = 0x745174  # GetAliasPath @ 005e6890 alias for save root
K1_CURRENTGAME: int = 0x73d8f4  # CURRENTGAME path ref
K1_FORMAT_PATH: int = 0x7454d4  # Format @ 005e5680 pattern for save path
K1_FORMAT_SCREENSHOT_PATTERN: int = 0x7454ac  # Format for screenshot path
K1_PATH_SKIP_SCREENSHOT_CMP: int = 0x73d71c  # operator!= @ 005e5390 compare path (skip screenshot if equal)


def get_free_disk_space_k1(path: Path) -> int:
    """GetFreeDiskSpace @ 004067b0 (CExoResMan).

    Returns free bytes on the filesystem for the given path.
    Callees in binary: CStr 005e5670, ResolveFileName 005e68a0, GetLength 005e5790,
    GetDiskFreeSpaceExA (external). We use shutil.disk_usage.
    """
    stat = shutil.disk_usage(path)
    return stat.free


def get_directory_size_k1(path: Path) -> int:
    """GetDirectorySize @ 005e6650.

    Returns total size in bytes of all files under path (recursive).
    Engine compares result with [EDI+0x100c8]. Returns 0 if path does not exist or is not a directory.
    """
    if not path.is_dir():
        return 0
    total = 0
    try:
        for p in path.rglob("*"):
            if p.is_file():
                total += p.stat().st_size
    except OSError:
        return 0
    return total


def clean_directory_k1(path: Path) -> bool:
    """CleanDirectory @ 00409460 (CExoResMan).

    Removes all contents under path so the directory is empty. Does not remove path itself.
    Single callee in binary: WipeDirectory @ 00408e90.
    Used on CreateDirectory2 failure path: engine retries create after CleanDirectory.
    Returns True if all contents were removed (or dir was already empty).
    """
    if not path.is_dir():
        return True
    try:
        for p in path.iterdir():
            if p.is_file():
                p.unlink()
            else:
                shutil.rmtree(p)
        return True
    except OSError:
        return False


def create_directory_k1(path: Path) -> bool:
    """CreateDirectory2 @ 004068c0 (CExoResMan).

    Creates directory and parents. Returns True if created or already exists.
    Callees: CStr 005e5670, ResolveFileName 005e68a0, CreateDirectoryA (external).
    """
    try:
        path.mkdir(parents=True, exist_ok=True)
        return True
    except OSError:
        return False


def has_enough_disk_space_for_save_game_k1(path: Path) -> bool:
    """HasEnoughDiskSpaceForSaveGame @ 004b2520.

    Callees: GetFreeDiskSpace @ 004067b0, ~CExoString 005e5c20, CExoString 005e5a90.
    Condition: (free >> 14) >= 0x640. Returns True if sufficient space.
    """
    free = get_free_disk_space_k1(path)
    return (free >> 14) >= 0x640


def get_alias_path_k1(alias_id: int, base_path: Path) -> Path:
    """GetAliasPath @ 005e6890.

    Returns resolved path for alias_id. Engine resolves from config; we map
    K1_ALIAS_SAVE_PATH (0x745174) to base_path. Callees in binary: config lookup,
    CExoString 005b3190/005e5a90, ResolveFileName 005e68a0.
    """
    if alias_id == K1_ALIAS_SAVE_PATH:
        return base_path.resolve()
    return base_path.resolve()


def format_path_k1(pattern_id: int, currentgame_path: Path, save_name: str) -> Path:
    """Format @ 005e5680.

    Engine uses pattern at pattern_id (0x7454d4 for path, 0x7454ac for screenshot)
    to build path from CURRENTGAME/MODULES and save name. Callees: CStr 005e5670,
    CExoString 005b3190, pattern format. We implement as currentgame_path / save_name.
    """
    if pattern_id == K1_FORMAT_PATH:
        return (currentgame_path / save_name).resolve()
    if pattern_id == K1_FORMAT_SCREENSHOT_PATTERN:
        return (currentgame_path / save_name).resolve()
    return (currentgame_path / save_name).resolve()


def exo_string_operator_plus_k1(prefix: Path, suffix: str) -> Path:
    """operator+ @ 005e5d10 (CExoString path concatenation).

    Engine concatenates CExoString path segments. Callees: CExoString 005b3190,
    GetLength 005e5790. We return prefix / suffix as Path.
    """
    return (prefix / suffix).resolve()


def cstr_from_path_k1(path: Path) -> str:
    """CStr @ 005e5670.

    Converts path (CExoString) to char* for C APIs. We return path as posix string.
    """
    return path.as_posix()


def build_full_save_path_k1(save_root: Path, save_name: str) -> Path:
    """Step 4: Build full path. GetAliasPath 005e6890, Format 005e5680, operator+ 005e5d10.

    Sequence: GetAliasPath(K1_ALIAS_SAVE_PATH) → base; Format(K1_FORMAT_PATH, base, save_name)
    with CURRENTGAME 0x73d8f4 and MODULES; operator+ for path segments. Returns full save path.
    """
    alias_path = get_alias_path_k1(K1_ALIAS_SAVE_PATH, save_root)
    full_path = format_path_k1(K1_FORMAT_PATH, alias_path, save_name)
    return full_path


def send_server_to_player_save_load_status_k1(
    type_val: int,
    subtype: int,
    player: Any,
) -> None:
    """SendServerToPlayerSaveLoad_Status @ 0056cb80.

    Sends save/load status to client. Callees: GetSWSMessage 004aedb0 (get message
    buffer), then fill save status struct: type 1, subtype 2, byte 1; at +0x18 and +0x20
    two CExoString fields; +0x8 and +0xc zeroed; +0x38 zero (see SAVE_LOAD_ENGINE_BEHAVIOR.md).
    Engine writes [EAX], [EAX+4], [EAX+0x10], [EAX+0x38], [EAX+0x18], [EAX+0x20], [EAX+8], [EAX+0xc].
    We build the struct and perform the send (no-op in Python; no real CSWSPlayer).
    """
    # Struct layout: type, subtype, byte; +0x8, +0xc zeroed; +0x18, +0x20 CExoString; +0x38 zero
    _type = type_val  # 1
    _subtype = subtype  # 2
    _byte = 1
    _zero_8 = 0
    _zero_c = 0
    _str_18 = ""
    _str_20 = ""
    _zero_38 = 0
    # Send to player (engine would call GetSWSMessage 004aedb0 and network send; we no-op)
    _ = (player, _type, _subtype, _byte, _zero_8, _zero_c, _str_18, _str_20, _zero_38)


def send_server_to_player_load_bar_start_stall_event_k1(event_type: int) -> None:
    """SendServerToPlayerLoadBar_StartStallEvent @ 0056c8b0.

    Notifies client to show load bar / stall event. Callees in binary: message build,
    send to client. We perform the call with exact parameter (event_type=2); no-op in Python.
    """
    _ = event_type


def _do_save_game_screenshot_k1(
    screenshot_path: Path,
    screenshot_data: Optional[bytes],
) -> None:
    """DoSaveGameScreenShot @ 00401080.

    Signature: void __stdcall DoSaveGameScreenShot(CExoString * param_1, bool param_2, bool param_3).
    Callees: CStr 005e5670, AurSaveGameSnapshot 00420f20(path, 1, 1).
    We write screenshot bytes to screenshot_path; engine uses AurSaveGameSnapshot(path, 1, 1).
    """
    if screenshot_data is None:
        return
    screenshot_path.parent.mkdir(parents=True, exist_ok=True)
    screenshot_path.write_bytes(screenshot_data)


def run_k1_save_flow(
    entry: SaveFolderEntry,
    *,
    min_free_bytes: int = K1_MIN_FREE_BYTES_ENGINE,
    required_save_bytes: Optional[int] = None,
    required_max_directory_bytes: Optional[int] = None,
    skip_screenshot_if_path_equal: Optional[str] = None,
    retry_create_after_clean: bool = True,
    write_components: bool = True,
) -> int:
    """SaveGame @ 004b58a0 (CServerExoAppInternal).

    int __thiscall SaveGame(CServerExoAppInternal * this, ulong param_1, CExoString * param_2, CSWSPlayer * param_3).
    Callers: SaveGame @ 004ae6e0, MainLoop @ 004babb0.
    Returns 1 on success, 0 on failure.

    Exhaustive sequence (every step from disassembly; offsets and labels exact):

    1. Get free disk space: GetFreeDiskSpace @ 004067b0. If free < min (HasEnoughDiskSpaceForSaveGame @ 004b2520
       threshold) or free < required_save_bytes ([EDI+0x100c8] when used), return 0.

    2. Create directory: CreateDirectory2 @ 004068c0. On failure: if retry_create_after_clean,
       CleanDirectory @ 00409460 then retry CreateDirectory2; else error path (return 0).

    3. GetDirectorySize @ 005e6650 vs [EDI+0x100c8]: if required_max_directory_bytes set and
       directory size >= it, return 0.

    4. Build full path: GetAliasPath 005e6890 (K1_ALIAS_SAVE_PATH 0x745174), Format 005e5680
       (K1_FORMAT_PATH 0x7454d4) with CURRENTGAME 0x73d8f4 and MODULES, operator+ 005e5d10,
       CStr 005e5670. Implemented by build_full_save_path_k1 and cstr_from_path_k1.

    5. SendServerToPlayerSaveLoad_Status(1, 2, player) @ 0056cb80: save status struct (type 1,
       subtype 2, byte 1; +0x8,+0xc zeroed; +0x18,+0x20 CExoString; +0x38 zero). Callees:
       GetSWSMessage 004aedb0. Implemented by send_server_to_player_save_load_status_k1(1, 2, None).

    6. SendServerToPlayerLoadBar_StartStallEvent(2) @ 0056c8b0. Implemented by
       send_server_to_player_load_bar_start_stall_event_k1(2).

    7. GetAliasPath @ 005e6890; operator!= @ 005e5390 for path 0x73d71c: if path equals
       skip_screenshot_if_path_equal, skip screenshot (engine skips DoSaveGameScreenShot when alias match).

    8. GetModule @ 005ed530, SetCameraForScreenShot @ 00638e80 — no-op in Python.

    9. DoSaveGameScreenShot @ 00401080: CStr, then AurSaveGameSnapshot 00420f20(path, 1, 1).
       We call _do_save_game_screenshot_k1 (write screenshot unless skipped).

    10. If param_1 != 0: GetSWGuiManager 005eda40, Draw 0040cc50, UpdateScreen 00401c10 — no-op in Python.

    11. Clear [this+0x1b930], [this+0x1b938] — no-op (no persistent CServerExoAppInternal in Python).

    12. Return 1. Optionally write SAVENFO, PARTYTABLE, GLOBALVARS, SAVEGAME.sav in engine order.
    """
    # Step 1: GetFreeDiskSpace @ 004067b0; compare with HasEnoughDiskSpaceForSaveGame @ 004b2520 / [EDI+0x100c8]
    save_path = entry.save_path
    root: Path = Path(save_path).parent
    path_obj: Path = Path(save_path)

    free: int = get_free_disk_space_k1(root)
    if free < min_free_bytes:
        return 0
    if required_save_bytes is not None and free < required_save_bytes:
        return 0

    # Step 2: CreateDirectory2 @ 004068c0; on failure CleanDirectory @ 00409460 then retry (engine error path)
    if not create_directory_k1(path_obj):
        if retry_create_after_clean:
            clean_directory_k1(path_obj)
            if not create_directory_k1(path_obj):
                return 0
        else:
            return 0

    # Step 3: GetDirectorySize @ 005e6650 vs [EDI+0x100c8]
    if required_max_directory_bytes is not None:
        dir_size = get_directory_size_k1(path_obj)
        if dir_size >= required_max_directory_bytes:
            return 0

    # Step 4: Build full path — GetAliasPath 005e6890, Format 005e5680 (0x7454d4), operator+ 005e5d10, CStr 005e5670
    save_name = path_obj.name
    path_obj = build_full_save_path_k1(root, save_name)
    _path_cstr = cstr_from_path_k1(path_obj)

    # Step 5: SendServerToPlayerSaveLoad_Status(1, 2, player) @ 0056cb80 — struct fill + send; callees GetSWSMessage 004aedb0
    send_server_to_player_save_load_status_k1(1, 2, None)

    # Step 6: SendServerToPlayerLoadBar_StartStallEvent(2) @ 0056c8b0
    send_server_to_player_load_bar_start_stall_event_k1(2)

    # Step 7: GetAliasPath @ 005e6890; operator!= @ 005e5390 (path K1_PATH_SKIP_SCREENSHOT_CMP 0x73d71c) → skip screenshot if path matches
    skip_screenshot = (
        skip_screenshot_if_path_equal is not None
        and path_obj.resolve().as_posix() == Path(skip_screenshot_if_path_equal).resolve().as_posix()
    )

    # Steps 8–9: GetModule @ 005ed530, SetCameraForScreenShot @ 00638e80 no-op. DoSaveGameScreenShot @ 00401080 → AurSaveGameSnapshot 00420f20
    if not skip_screenshot:
        screenshot_path = path_obj / entry.SCREENSHOT_NAME.resname
        _do_save_game_screenshot_k1(screenshot_path, entry.screenshot)

    # Step 10: param_1 != 0 → Draw, UpdateScreen no-op. Step 11: [this+0x1b930], [this+0x1b938] no-op.

    # Step 12: Write components in engine order (SAVENFO, PARTYTABLE, GLOBALVARS, SAVEGAME.sav)
    if write_components:
        entry.save_info.save()
        entry.partytable.save()
        entry.globals.save()
        entry.sav.save()

    return 1


def run_k1_load_flow(entry: SaveFolderEntry) -> Any:
    """LoadGame @ 004ba640 (CServerExoAppInternal).

    int __thiscall LoadGame(CServerExoAppInternal * this, ulong param_1, CExoString * param_2, CExoString * param_3, CSWSPlayer * param_4).
    Caller: LoadGame @ 004ae6f0.
    Returns load result (entry).

    Exhaustive sequence (every step from disassembly; offsets and labels exact):

    1. CExoString locals — implementation detail.

    2. SetLoadBarProgress(1, 0xa) @ 005edd40 — no-op in Python.

    3. SetLoadStep @ 005edfd0: (0xa,0), (0x14,1), (0x17,2), (0x17,3), (0x17,4) — no-op in Python.

    4. Build path: param_1 != -1 → Format 0x7455ac else operator=; alias 0x745174, operator+ — we use entry.save_path.

    5. AddResourceDirectory @ 00408800(path) — in Python we open files from path.

    6. LoadTableInfo @ 00565d20 at [this+0x1b770] — we load PARTYTABLE (table cache).

    7. Load @ 0052ade0 at [this+0x100fc] — we load GLOBALVARS and core data.

    8. RemoveResourceDirectory @ 004088d0 — no-op in Python.

    9. Set [this+0x1b87c]=1, [this+0x1b930]=0, [this+0x1b938]=0 — no-op.

    10. LoadModule @ 004b95b0(this, param_3, param_4) — we load SAVEGAME.sav (nested capsule).

    11. Return result (entry).
    """
    # Steps 2–3: SetLoadBarProgress 005edd40, SetLoadStep 005edfd0 no-op.

    # Step 4: Path is entry.save_path (alias 0x745174, Format 0x7455ac).

    # Step 5: AddResourceDirectory @ 00408800 — we use entry.save_path for file reads.

    # Step 6: LoadTableInfo @ 00565d20 at [this+0x1b770]
    entry.partytable.load()

    # Step 7: Load @ 0052ade0 at [this+0x100fc]
    entry.save_info.load()
    entry.globals.load()

    # Step 8: RemoveResourceDirectory @ 004088d0 no-op.

    # Step 9: [this+0x1b87c]=1, [this+0x1b930]=0, [this+0x1b938]=0 no-op.

    # Step 10: LoadModule @ 004b95b0
    entry.sav.load()

    # Screenshot (engine may load elsewhere; we load for fidelity)
    screenshot_path: Path = Path(entry.save_path) / entry.SCREENSHOT_NAME.resname
    if screenshot_path.is_file():
        entry.screenshot = screenshot_path.read_bytes()
    else:
        entry.screenshot = None

    return entry


def run_save_flow(
    entry: SaveFolderEntry,
    game_install_path: str | Path,
    **kwargs: Any,
) -> int:
    """Run the engine-identical save flow for the game at game_install_path.

    Uses determine_game(game_install_path) from pykotor.tools.heuristics to decide
    K1 (SaveGame @ 004b58a0, k1_win_gog_swkotor.exe) vs K2/TSL (StallEventSaveGame @ 007b7980,
    k2_win_gog_aspyr_swkotor2.exe), then calls run_k1_save_flow or run_tsl_save_flow.
    Extra kwargs are passed through (K1 accepts all; TSL only min_free_bytes, write_components).

    Returns:
        1 on success, 0 on failure. Returns 0 if game cannot be determined.
    """
    game = determine_game(game_install_path)
    if game is None:
        return 0
    if game in _GAMES_K1:
        return run_k1_save_flow(entry, **kwargs)
    if game in _GAMES_K2_TSL:
        tsl_kw = {k: v for k, v in kwargs.items() if k in ("min_free_bytes", "write_components")}
        return run_tsl_save_flow(entry, **tsl_kw)
    return 0


def run_load_flow(
    entry: SaveFolderEntry,
    game_install_path: str | Path,
) -> Any:
    """Run the engine-identical load flow for the game at game_install_path.

    Uses determine_game(game_install_path) from pykotor.tools.heuristics to decide
    K1 (LoadGame @ 004ba640) vs K2/TSL (LoadGame flow FUN_007b2f00, LoadModule 007b4fd0),
    then calls run_k1_load_flow or run_tsl_load_flow.

    Returns:
        The loaded SaveFolderEntry. Raises ValueError if game cannot be determined.
    """
    game = determine_game(game_install_path)
    if game is None:
        raise ValueError(f"Cannot determine game for path: {game_install_path}")
    if game in _GAMES_K1:
        return run_k1_load_flow(entry)
    if game in _GAMES_K2_TSL:
        return run_tsl_load_flow(entry)
    raise ValueError(f"Unsupported game for save/load flow: {game!r}")
